import os
import os.path
import io
import unittest
import unittest.mock
import datetime

from libpyvinyl.Parameters.Collections import CalculatorParameters

from mcstasscript.interface.instr import McStas_instr
from mcstasscript.interface.instr import McXtrace_instr
from mcstasscript.helper.formatting import bcolors
from mcstasscript.tests.helpers_for_tests import WorkInTestDir
from mcstasscript.helper.exceptions import McStasError
from mcstasscript.helper.mcstas_objects import Component
from mcstasscript.helper.beam_dump_database import BeamDump

run_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '.')

class DummyComponent(Component):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def set_parameters(self, *args, **kwargs):
        pass

def setup_instr_no_path():
    """
    Sets up a neutron instrument without a package_path
    """

    with WorkInTestDir() as handler:
        instrument = McStas_instr("test_instrument")

    return instrument


def setup_x_ray_instr_no_path():
    """
    Sets up a X-ray instrument without a package_path
    """

    with WorkInTestDir() as handler:
        instrument = McXtrace_instr("test_instrument")

    return instrument


def setup_instr_root_path():
    """
    Sets up a neutron instrument with root package_path
    """
    with WorkInTestDir() as handler:
        instrument = McStas_instr("test_instrument", package_path="/")

    return instrument


def setup_x_ray_instr_root_path():
    """
    Sets up a X-ray instrument with root package_path
    """
    with WorkInTestDir() as handler:
        instrument = McXtrace_instr("test_instrument", package_path="/")

    return instrument


def setup_instr_with_path():
    """
    Sets up an instrument with a valid package_path, but it points to
    the dummy installation in the test folder.
    """

    with WorkInTestDir() as handler:
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        dummy_path = os.path.join(THIS_DIR, "dummy_mcstas")
        instrument = McStas_instr("test_instrument",
                                  package_path=dummy_path, executable_path=dummy_path)

    return instrument


def setup_x_ray_instr_with_path():
    """
    Sets up an instrument with a valid package_path, but it points to
    the dummy installation in the test folder.
    """

    THIS_DIR = os.path.dirname(os.path.abspath(__file__))
    dummy_path = os.path.join(THIS_DIR, "dummy_mcstas")

    with WorkInTestDir() as handler:
        instrument = McXtrace_instr("test_instrument", package_path=dummy_path)

    return instrument


def setup_instr_with_input_path():
    """
    Sets up an instrument with a valid package_path, but it points to
    the dummy installation in the test folder. In addition the input_path
    is set to a folder in the test directory using an absolute path.
    """

    THIS_DIR = os.path.dirname(os.path.abspath(__file__))
    dummy_path = os.path.join(THIS_DIR, "dummy_mcstas")
    input_path = os.path.join(THIS_DIR, "test_input_folder")

    with WorkInTestDir() as handler:
        instrument = McStas_instr("test_instrument",
                                  package_path=dummy_path,
                                  input_path=input_path)

    return instrument


def setup_instr_with_input_path_relative():
    """
    Sets up an instrument with a valid package_path, but it points to
    the dummy installation in the test folder. In addition the input_path
    is set to a folder in the test directory using a relative path.
    """

    with WorkInTestDir() as handler:
        instrument = McStas_instr("test_instrument",
                                  package_path="dummy_mcstas",
                                  input_path="test_input_folder")

    return instrument


def setup_populated_instr():
    """
    Sets up a neutron instrument with some features used and three components
    """
    instr = setup_instr_root_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    instr.add_component("first_component", "test_for_reading")
    instr.add_component("second_component", "test_for_reading")
    instr.add_component("third_component", "test_for_reading")

    return instr


def setup_populated_instr_with_dummy_MCPL_comps():
    """
    Sets up a neutron instrument with some features used and three components
    """
    instr = setup_populated_instr()

    instr.component_class_lib["MCPL_input"] = DummyComponent
    instr.component_class_lib["MCPL_output"] = DummyComponent

    return instr

def setup_populated_instr_with_dummy_path():
    """
    Sets up a neutron instrument with some features used and three components

    Here uses the dummy mcstas installation as path and sets required
    parameters so that a run is possible.
    """
    instr = setup_instr_with_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    comp1 = instr.add_component("first_component", "test_for_reading")
    comp1.gauss = 1.2
    comp1.test_string = "a_string"
    comp2 = instr.add_component("second_component", "test_for_reading")
    comp2.gauss = 1.4
    comp2.test_string = "b_string"
    comp3 = instr.add_component("third_component", "test_for_reading")
    comp3.gauss = 1.6
    comp3.test_string = "c_string"

    return instr


def setup_populated_x_ray_instr():
    """
    Sets up a X-ray instrument with some features used and three components
    """
    instr = setup_x_ray_instr_root_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    instr.add_component("first_component", "test_for_reading")
    instr.add_component("second_component", "test_for_reading")
    instr.add_component("third_component", "test_for_reading")

    return instr


def setup_populated_x_ray_instr_with_dummy_path():
    """
    Sets up a x-ray instrument with some features used and three components

    Here uses the dummy mcstas installation as path and sets required
    parameters so that a run is possible.
    """
    instr = setup_x_ray_instr_with_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    comp1 = instr.add_component("first_component", "test_for_reading")
    comp1.gauss = 1.2
    comp1.test_string = "a_string"
    comp2 = instr.add_component("second_component", "test_for_reading")
    comp2.gauss = 1.4
    comp2.test_string = "b_string"
    comp3 = instr.add_component("third_component", "test_for_reading")
    comp3.gauss = 1.6
    comp3.test_string = "c_string"

    return instr


def setup_populated_with_some_options_instr():
    """
    Sets up a neutron instrument with some features used and two components
    """
    instr = setup_instr_root_path()

    instr.add_parameter("double", "theta")
    instr.add_parameter("double", "has_default", value=37)
    instr.add_declare_var("double", "two_theta")
    instr.append_initialize("two_theta = 2.0*theta;")

    comp1 = instr.add_component("first_component", "test_for_reading")
    comp1.set_AT([0, 0, 1])
    comp1.set_GROUP("Starters")
    comp2 = instr.add_component("second_component", "test_for_reading")
    comp2.set_AT([0, 0, 2], RELATIVE="first_component")
    comp2.set_ROTATED([0, 30, 0])
    comp2.set_WHEN("1==1")
    comp2.yheight = 1.23
    instr.add_component("third_component", "test_for_reading")

    return instr


def insert_mock_dump(instr, component_name, run_name="Run", tag=0):
    dump = BeamDump("", {}, component_name, run_name, tag=tag)
    dump.file_present = lambda *_: True  # Overwrite file check

    instr.dump_database.data[component_name] = {}
    instr.dump_database.data[component_name][run_name] = {}
    instr.dump_database.data[component_name][run_name][tag] = dump


class TestMcStas_instr(unittest.TestCase):
    """
    Tests of the main class in McStasScript called McStas_instr.
    """

    def test_simple_initialize(self):
        """
        Test basic initialization runs
        """
        my_instrument = setup_instr_root_path()

        self.assertEqual(my_instrument.name, "test_instrument")

    def test_complex_initialize(self):
        """
        Tests all keywords work in initialization
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        my_instrument = McStas_instr("test_instrument",
                                     author="Mads",
                                     origin="DMSC",
                                     executable_path="./dummy_mcstas/contrib",
                                     package_path="./dummy_mcstas/misc")

        os.chdir(current_work_dir)

        self.assertEqual(my_instrument.author, "Mads")
        self.assertEqual(my_instrument.origin, "DMSC")
        self.assertEqual(my_instrument._run_settings["executable_path"],
                         "./dummy_mcstas/contrib")
        self.assertEqual(my_instrument._run_settings["package_path"],
                         "./dummy_mcstas/misc")

    def test_load_config_file(self):
        """
        Test that configuration file is read correctly. In order to have
        an independent test, the yaml file is read manually instead of
        using the yaml package.
        """
        # Load configuration file and read manually
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        configuration_file_name = os.path.join(THIS_DIR,
                                               "..", "configuration.yaml")

        if not os.path.isfile(configuration_file_name):
            raise NameError("Could not find configuration file!")

        f = open(configuration_file_name, "r")

        lines = f.readlines()
        for line in lines:
            line = line.strip()
            if line.startswith("mcrun_path:"):
                parts = line.split(" ")
                correct_mcrun_path = parts[1]

            if line.startswith("mcstas_path:"):
                parts = line.split(" ")
                correct_mcstas_path = parts[1]

            if line.startswith("characters_per_line:"):
                parts = line.split(" ")
                correct_n_of_characters = int(parts[1])

        f.close()

        # Check the value matches what is loaded by initialization
        my_instrument = setup_instr_no_path()

        self.assertEqual(my_instrument._run_settings["executable_path"], correct_mcrun_path)
        self.assertEqual(my_instrument._run_settings["package_path"], correct_mcstas_path)
        self.assertEqual(my_instrument.line_limit, correct_n_of_characters)

    def test_load_config_file_x_ray(self):
        """
        Test that configuration file is read correctly. In order to have
        an independent test, the yaml file is read manually instead of
        using the yaml package.
        """
        # Load configuration file and read manually
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        configuration_file_name = os.path.join(THIS_DIR,
                                               "..", "configuration.yaml")

        if not os.path.isfile(configuration_file_name):
            raise NameError("Could not find configuration file!")

        f = open(configuration_file_name, "r")

        lines = f.readlines()
        for line in lines:
            line = line.strip()
            if line.startswith("mxrun_path:"):
                parts = line.split(" ")
                correct_mxrun_path = parts[1]

            if line.startswith("mcxtrace_path:"):
                parts = line.split(" ")
                correct_mcxtrace_path = parts[1]

            if line.startswith("characters_per_line:"):
                parts = line.split(" ")
                correct_n_of_characters = int(parts[1])

        f.close()

        # Check the value matches what is loaded by initialization
        my_instrument = setup_x_ray_instr_no_path()

        self.assertEqual(my_instrument._run_settings["executable_path"], correct_mxrun_path)
        self.assertEqual(my_instrument._run_settings["package_path"], correct_mcxtrace_path)
        self.assertEqual(my_instrument.line_limit, correct_n_of_characters)

    def test_load_libpyvinyl_parameters(self):
        parameters = CalculatorParameters()
        int_par = parameters.new_parameter("int_parameter", comment="integer parameter")
        int_par.value = 3

        double_par = parameters.new_parameter("double_parameter", unit="meV")
        double_par.value = 3.0

        string_par = parameters.new_parameter("string_parameter")
        string_par.value = "hello world"

        secret_par = parameters.new_parameter("no_value_par")

        instr = McStas_instr("test_instr", parameters=parameters)

        self.assertEqual(int_par.type, "double")
        self.assertEqual(double_par.type, "double")
        self.assertEqual(string_par.type, "string")
        self.assertEqual(secret_par.type, None)

        instr.set_parameters(int_parameter=4)
        self.assertEqual(int_par.value, 4)

        int_par.value = 5
        self.assertEqual(instr.parameters["int_parameter"].value, 5)

    def test_simple_add_parameter(self):
        """
        This is just an interface to a function that is tested
        elsewhere, so only a basic test is performed here.

        ParameterVariable is tested in test_parameter_variable.
        """
        instr = setup_instr_root_path()

        parameter = instr.add_parameter("double", "theta", comment="test par")

        self.assertEqual(parameter.name, "theta")
        self.assertEqual(parameter.comment, "test par")
        self.assertTrue(parameter in instr.parameters.parameters.values())

    def test_user_var_block_add_parameter(self):
        """
        Checks that adding a parameter with a name already used for a
        user variable fails with NameError
        """
        instr = setup_instr_root_path()

        instr.add_user_var("double", "theta")

        with self.assertRaises(NameError):
            instr.add_parameter("double", "theta", comment="test par")


    def test_declare_var_block_add_parameter(self):
        """
        Checks that adding a parameter with a name already used for a
        declared variable fails with NameError
        """
        instr = setup_instr_root_path()

        instr.add_declare_var("double", "theta")

        with self.assertRaises(NameError):
            instr.add_parameter("double", "theta", comment="test par")

    def test_infer_name_add_parameter(self):
        """
        Test that name can be ommited when defining a parameter and that
        the python variable name is used in its place.
        """
        instr = setup_instr_root_path()

        theta = instr.add_parameter(type="double", comment="test par")

        self.assertEqual(theta.name, "theta")
        self.assertEqual(theta.comment, "test par")
        self.assertTrue(theta in instr.parameters.parameters.values())

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_show_parameters(self, mock_stdout):
        """
        Testing that parameters are displayed correctly
        """
        instr = setup_instr_root_path()

        instr.add_parameter("theta", comment="test par")
        instr.add_parameter("double", "par_double", comment="test par")
        instr.add_parameter("int", "int_par", value=8, comment="test par")
        instr.add_parameter("int", "slits", comment="test par")
        instr.add_parameter("string", "ref",
                            value="string", comment="new string")

        instr.show_parameters(line_length=300)

        output = mock_stdout.getvalue().split("\n")

        self.assertEqual(output[0], "       theta                 // test par")
        self.assertEqual(output[1], "double par_double            // test par")
        self.assertEqual(output[2], "int    int_par     = 8       // test par")
        self.assertEqual(output[3], "int    slits                 // test par")
        self.assertEqual(output[4], "string ref         = string  // new string")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_show_parameters_line_break(self, mock_stdout):
        """
        Testing that parameters are displayed correctly

        Here multiple lines are used for a long comment that was
        dynamically broken up.
        """
        instr = setup_instr_root_path()

        instr.add_parameter("theta", comment="test par")
        instr.add_parameter("double", "par_double", comment="test par")
        instr.add_parameter("int", "int_par", value=8, comment="test par")
        instr.add_parameter("int", "slits", comment="test par")
        instr.add_parameter("string", "ref",
                            value="string", comment="new string")

        longest_comment = ("This is a very long comment meant for "
                           + "testing the dynamic line breaking "
                           + "that is used in this method. It needs "
                           + "to have many lines in order to ensure "
                           + "it really works.")

        instr.add_parameter("double", "value",
                            value="37", comment=longest_comment)

        instr.show_parameters(line_length=80)

        output = mock_stdout.getvalue().split("\n")

        self.assertEqual(output[0], "       theta                 // test par")
        self.assertEqual(output[1], "double par_double            // test par")
        self.assertEqual(output[2], "int    int_par     = 8       // test par")
        self.assertEqual(output[3], "int    slits                 // test par")
        self.assertEqual(output[4], "string ref         = string  // new string")
        comment_line = "This is a very long comment meant for "
        self.assertEqual(output[5], "double value       = 37      // "
                                    + comment_line)
        comment_line = "testing the dynamic line breaking that is "
        self.assertEqual(output[6], " "*33 + comment_line)
        comment_line = "used in this method. It needs to have many "
        self.assertEqual(output[7], " "*33 + comment_line)
        comment_line = "lines in order to ensure it really works. "
        self.assertEqual(output[8], " "*33 + comment_line)

    def test_simple_add_declare_variable(self):
        """
        This is just an interface to a function that is tested
        elsewhere, so only a basic test is performed here.

        DeclareVariable is tested in test_declare_variable.
        """
        instr = setup_instr_root_path()

        instr.add_declare_var("double", "two_theta", comment="test par")

        self.assertEqual(instr.declare_list[0].name, "two_theta")
        self.assertEqual(instr.declare_list[0].comment, " // test par")

    def test_parameter_block_add_declare_variable(self):
        """
        Checks a NameError is raised when using declare variable of same
        name as instrument parameter.
        """
        instr = setup_instr_root_path()

        instr.add_parameter("two_theta")
        with self.assertRaises(NameError):
            instr.add_declare_var("double", "two_theta")

    def test_user_var_block_add_declare_variable(self):
        """
        Checks a NameError is raised when using declare variable of same
        name as declared variable.
        """
        instr = setup_instr_root_path()

        instr.add_user_var("double", "two_theta")
        with self.assertRaises(NameError):
            instr.add_declare_var("double", "two_theta")

    def test_infer_add_declare_variable(self):
        """
        Check that name can be inferred from python variable when
        adding a declared variable.
        """
        instr = setup_instr_root_path()

        two_theta = instr.add_declare_var("double", comment="test par")

        self.assertEqual(instr.declare_list[0].name, "two_theta")
        self.assertEqual(instr.declare_list[0].comment, " // test par")

    def test_simple_add_user_variable(self):
        """
        This is just an interface to a function that is tested
        elsewhere, so only a basic test is performed here.

        DeclareVariable is tested in test_declare_variable.
        """
        instr = setup_instr_root_path()

        user_var = instr.add_user_var("double", "two_theta_user", comment="test par")

        self.assertEqual(user_var.name, "two_theta_user")
        self.assertEqual(user_var.comment, " // test par")
        self.assertEqual(instr.user_var_list[0].name, "two_theta_user")
        self.assertEqual(instr.user_var_list[0].comment, " // test par")

        with self.assertRaises(ValueError):
            instr.add_user_var("double", "illegal", value=8)

    def test_declare_block_add_user_variable(self):
        """
        Checks a NameError is raised when using user variable of same
        name as declare variable already defined.
        """
        instr = setup_instr_root_path()

        instr.add_declare_var("double", "two_theta")
        with self.assertRaises(NameError):
            instr.add_user_var("double", "two_theta")

    def test_parameter_block_add_user_variable(self):
        """
        Checks a NameError is raised when using user variable of same
        name as parameter already defined.
        """
        instr = setup_instr_root_path()

        instr.add_parameter("two_theta")
        with self.assertRaises(NameError):
            instr.add_user_var("double", "two_theta")

    def test_simple_append_declare(self):
        """
        Appending to declare adds an object to the declare list, and the
        allowed types are either strings or DeclareVariable objects.
        Here only strings are added.
        """
        instr = setup_instr_root_path()

        instr.append_declare("First line of declare")
        instr.append_declare("Second line of declare")
        instr.append_declare("Third line of declare")

        self.assertEqual(instr.declare_list[0],
                         "First line of declare")
        self.assertEqual(instr.declare_list[1],
                         "Second line of declare")
        self.assertEqual(instr.declare_list[2],
                         "Third line of declare")

    def test_simple_append_declare_var_mix(self):
        """
        Appending to declare adds an object to the declare list, and the
        allowed types are either strings or DeclareVariable objects.
        Here a mix of strings and DeclareVariable objects are added.
        """
        instr = setup_instr_root_path()

        instr.append_declare("First line of declare")
        instr.add_declare_var("double", "two_theta", comment="test par")
        instr.append_declare("Third line of declare")

        self.assertEqual(instr.declare_list[0],
                         "First line of declare")
        self.assertEqual(instr.declare_list[1].name, "two_theta")
        self.assertEqual(instr.declare_list[1].comment, " // test par")
        self.assertEqual(instr.declare_list[2],
                         "Third line of declare")

    def test_simple_append_initialize(self):
        """
        The initialize section is held as a string. This method
        appends that string.
        """
        instr = setup_instr_root_path()

        self.assertEqual(instr.initialize_section,
                         "// Start of initialize for generated "
                         + "test_instrument\n")

        instr.append_initialize("First line of initialize")
        instr.append_initialize("Second line of initialize")
        instr.append_initialize("Third line of initialize")

        self.assertEqual(instr.initialize_section,
                         "// Start of initialize for generated "
                         + "test_instrument\n"
                         + "First line of initialize\n"
                         + "Second line of initialize\n"
                         + "Third line of initialize\n")

    def test_simple_append_initialize_no_new_line(self):
        """
        The initialize section is held as a string. This method
        appends that string without making a new line.
        """
        instr = setup_instr_root_path()

        self.assertEqual(instr.initialize_section,
                         "// Start of initialize for generated "
                         + "test_instrument\n")

        instr.append_initialize_no_new_line("A")
        instr.append_initialize_no_new_line("B")
        instr.append_initialize_no_new_line("CD")

        self.assertEqual(instr.initialize_section,
                         "// Start of initialize for generated "
                         + "test_instrument\n"
                         + "ABCD")

    def test_simple_append_finally(self):
        """
        The finally section is held as a string. This method
        appends that string.
        """
        instr = setup_instr_root_path()

        self.assertEqual(instr.finally_section,
                         "// Start of finally for generated "
                         + "test_instrument\n")

        instr.append_finally("First line of finally")
        instr.append_finally("Second line of finally")
        instr.append_finally("Third line of finally")

        self.assertEqual(instr.finally_section,
                         "// Start of finally for generated "
                         + "test_instrument\n"
                         + "First line of finally\n"
                         + "Second line of finally\n"
                         + "Third line of finally\n")

    def test_simple_append_finally_no_new_line(self):
        """
        The finally section is held as a string. This method
        appends that string without making a new line.
        """
        instr = setup_instr_root_path()

        self.assertEqual(instr.finally_section,
                         "// Start of finally for generated "
                         + "test_instrument\n")

        instr.append_finally_no_new_line("A")
        instr.append_finally_no_new_line("B")
        instr.append_finally_no_new_line("CD")

        self.assertEqual(instr.finally_section,
                         "// Start of finally for generated "
                         + "test_instrument\n"
                         + "ABCD")

    def test_simple_append_trace(self):
        """
        The trace section is held as a string. This method
        appends that string. Only used for writing c files, which is not
        the main way to use McStasScript.
        """
        instr = setup_instr_root_path()

        self.assertEqual(instr.trace_section,
                         "// Start of trace section for generated "
                         + "test_instrument\n")

        instr.append_trace("First line of trace")
        instr.append_trace("Second line of trace")
        instr.append_trace("Third line of trace")

        self.assertEqual(instr.trace_section,
                         "// Start of trace section for generated "
                         + "test_instrument\n"
                         + "First line of trace\n"
                         + "Second line of trace\n"
                         + "Third line of trace\n")

    def test_simple_append_trace_no_new_line(self):
        """
        The trace section is held as a string. This method appends that string
        without making a new line. Only used for writing c files, which is not
        the main way to use McStasScript.
        """
        instr = setup_instr_root_path()

        self.assertEqual(instr.trace_section,
                         "// Start of trace section for generated "
                         + "test_instrument\n")

        instr.append_trace_no_new_line("A")
        instr.append_trace_no_new_line("B")
        instr.append_trace_no_new_line("CD")

        self.assertEqual(instr.trace_section,
                         "// Start of trace section for generated "
                         + "test_instrument\n"
                         + "ABCD")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_available_components_simple(self, mock_stdout):
        """
        Simple test of show components to show component categories
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        instr = setup_instr_with_path()

        instr.available_components()

        os.chdir(current_work_dir)

        output = mock_stdout.getvalue()
        output = output.split("\n")

        self.assertEqual(output[0],
                         "The following components are found in the "
                         + "work directory / input_path:")
        self.assertEqual(output[1], "     test_for_reading.comp")
        self.assertEqual(output[2], "These definitions will be used "
                         + "instead of the installed versions.")
        self.assertEqual(output[3],
                         "Here are the available component categories:")
        self.assertEqual(output[4], " misc")
        self.assertEqual(output[5], " sources")
        self.assertEqual(output[6], " work directory")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_available_components_folder(self, mock_stdout):
        """
        Simple test of show components to show components in current work
        directory.
        """
        instr = setup_instr_with_path()

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        instr.available_components("work directory")

        os.chdir(current_work_dir)

        output = mock_stdout.getvalue()
        output = output.split("\n")

        self.assertEqual(output[0],
                         "The following components are found in the "
                         + "work directory / input_path:")
        self.assertEqual(output[1], "     test_for_reading.comp")
        self.assertEqual(output[2], "These definitions will be used "
                         + "instead of the installed versions.")
        self.assertEqual(output[3],
                         "Here are all components in the work directory "
                         + "category.")
        self.assertEqual(output[4], " test_for_reading")
        self.assertEqual(output[5], "")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_available_components_input_path_simple(self, mock_stdout):
        """
        Simple test of input_path being recognized and passed
        to component_reader so PSDlin_monitor is overwritten
        """
        instr = setup_instr_with_input_path()

        instr.available_components()

        output = mock_stdout.getvalue()
        output = output.split("\n")

        self.assertEqual(output[0],
                         "The following components are found in the "
                         + "work directory / input_path:")
        self.assertEqual(output[1], "     test_for_structure.comp")
        self.assertEqual(output[2], "These definitions will be used "
                         + "instead of the installed versions.")
        self.assertEqual(output[3],
                         "Here are the available component categories:")
        self.assertEqual(output[4], " misc")
        self.assertEqual(output[5], " sources")
        self.assertEqual(output[6], " work directory")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_available_components_input_path_custom(self, mock_stdout):
        """
        Simple test of input_path being recognized and passed
        to component_reader so PSDlin_monitor is overwritten
        Here dummy_mcstas and input_path are set using relative
        paths instead of absolute paths.
        """
        instr = setup_instr_with_input_path_relative()

        instr.available_components()

        output = mock_stdout.getvalue()
        output = output.split("\n")

        self.assertEqual(output[0],
                         "The following components are found in the "
                         + "work directory / input_path:")
        self.assertEqual(output[1], "     test_for_structure.comp")
        self.assertEqual(output[2], "These definitions will be used "
                         + "instead of the installed versions.")
        self.assertEqual(output[3],
                         "Here are the available component categories:")
        self.assertEqual(output[4], " misc")
        self.assertEqual(output[5], " sources")
        self.assertEqual(output[6], " work directory")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_component_help(self, mock_stdout):
        """
        Simple test of component help
        """
        instr = setup_instr_with_path()

        instr.component_help("test_for_reading", line_length=90)
        # This call creates a dummy component and calls its
        # show_parameter method which has been tested. Here we
        # need to ensure the call is successful, not test all
        # output from the call.

        output = mock_stdout.getvalue()
        output = output.split("\n")

        self.assertEqual(output[3], " ___ Help test_for_reading " + "_"*63)

        legend = ("|"
                  + bcolors.BOLD + "optional parameter" + bcolors.ENDC
                  + "|"
                  + bcolors.BOLD + bcolors.UNDERLINE
                  + "required parameter"
                  + bcolors.ENDC + bcolors.ENDC
                  + "|"
                  + bcolors.BOLD + bcolors.OKBLUE
                  + "default value"
                  + bcolors.ENDC + bcolors.ENDC
                  + "|"
                  + bcolors.BOLD + bcolors.OKGREEN
                  + "user specified value"
                  + bcolors.ENDC + bcolors.ENDC
                  + "|")

        self.assertEqual(output[4], legend)

        par_name = bcolors.BOLD + "radius" + bcolors.ENDC
        value = (bcolors.BOLD + bcolors.OKBLUE
                 + "0.1" + bcolors.ENDC + bcolors.ENDC)
        comment = ("// Radius of circle in (x,y,0) plane where "
                   + "neutrons are generated.")
        self.assertEqual(output[5],
                         par_name + " = " + value + " [m] " + comment)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_create_component_instance_simple(self, mock_stdout):
        """
        Tests successful use of _create_component_instance

        _create_component_instance will make a dynamic subclass of
        component with the information from the component files read
        from disk.  The subclass is saved in a dict for reuse in
        case the same component type is requested again.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        instr = setup_instr_with_path()

        os.chdir(current_work_dir)

        comp = instr._create_component_instance("test_component",
                                                "test_for_reading")

        self.assertEqual(comp.radius, None)
        self.assertIn("radius", comp.parameter_names)
        self.assertEqual(comp.parameter_defaults["radius"], 0.1)
        self.assertEqual(comp.parameter_types["radius"], "double")
        self.assertEqual(comp.parameter_units["radius"], "m")

        comment = ("Radius of circle in (x,y,0) plane where "
                   + "neutrons are generated.")
        self.assertEqual(comp.parameter_comments["radius"], comment)
        self.assertEqual(comp.category, "work directory")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_create_component_instance_complex(self, mock_stdout):
        """
        Tests successful use of _create_component_instance while using
        keyword arguments in creation

        _create_component_instance will make a dynamic subclass of
        component with the information from the component files read
        from disk.  The subclasses is saved in a dict for reuse in
        case the same component type is requested again.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        instr = setup_instr_with_path()

        os.chdir(current_work_dir)

        # Setting relative to home, should be passed to component
        comp = instr._create_component_instance("test_component",
                                                "test_for_reading",
                                                RELATIVE="home")

        self.assertEqual(comp.radius, None)
        self.assertIn("radius", comp.parameter_names)
        self.assertEqual(comp.parameter_defaults["radius"], 0.1)
        self.assertEqual(comp.parameter_types["radius"], "double")
        self.assertEqual(comp.parameter_units["radius"], "m")

        comment = ("Radius of circle in (x,y,0) plane where "
                   + "neutrons are generated.")
        self.assertEqual(comp.parameter_comments["radius"], comment)
        self.assertEqual(comp.category, "work directory")

        # The keyword arguments of the call should be passed to the
        # new instance of the component. This is checked by reading
        # the relative attributes which were set to home in the call
        self.assertEqual(comp.AT_relative, "RELATIVE home")
        self.assertEqual(comp.ROTATED_relative, "RELATIVE home")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_add_component_simple(self, mock_stdout):
        """
        Testing add_component in simple case.

        The add_component method adds a new component object to the
        instrument and keeps track of its location within the
        sequence of components.  Normally a new component is added to
        the end of the sequence, but the before and after keywords can
        be used to select another location.
        """

        instr = setup_instr_with_path()

        comp = instr.add_component("test_component", "test_for_reading")

        self.assertEqual(len(instr.component_list), 1)
        self.assertEqual(instr.component_list[0].name, "test_component")

        # Test the resulting object functions as intended
        comp.set_GROUP("developers")
        self.assertEqual(instr.component_list[0].GROUP, "developers")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_add_component_infer_name(self, mock_stdout):
        """
        Testing add_component works when name is left out and inferred
        from the name of the python variable in the call.
        """

        instr = setup_instr_with_path()

        test_component = instr.add_component("test_for_reading")

        self.assertEqual(len(instr.component_list), 1)
        self.assertEqual(instr.component_list[0].name, "test_component")

        # Test the resulting object functions as intended
        test_component.set_GROUP("developers")
        self.assertEqual(instr.component_list[0].GROUP, "developers")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_add_component_simple_keyword(self, mock_stdout):
        """
        Testing add_component with keyword argument for the component

        The add_component method adds a new component object to the
        instrument and keeps track of its location within the
        sequence of components.  Normally a new component is added to
        the end of the sequence, but the before and after keywords can
        be used to select another location. Here keyword passing is
        tested.
        """

        instr = setup_instr_with_path()

        instr.add_component("test_component",
                            "test_for_reading",
                            WHEN="1<2")

        self.assertEqual(len(instr.component_list), 1)
        self.assertEqual(instr.component_list[0].name, "test_component")
        self.assertEqual(instr.component_list[0].component_name,
                         "test_for_reading")

        self.assertEqual(instr.component_list[0].WHEN, "WHEN (1<2)")

    def test_add_component_simple_before(self):
        """
        Testing add_component with before keyword argument for the method

        The add_component method adds a new component object to the
        instrument and keeps track of its location within the
        sequence of components.  Normally a new component is added to
        the end of the sequence, but the before and after keywords can
        be used to select another location, here before is tested.
        """

        instr = setup_populated_instr()

        instr.add_component("test_component",
                            "test_for_reading",
                            before="first_component")

        self.assertEqual(len(instr.component_list), 4)
        self.assertEqual(instr.component_list[0].name, "test_component")
        self.assertEqual(instr.component_list[3].name, "third_component")

    def test_add_component_simple_after(self):
        """
        Testing add_component with after keyword argument for the method

        The add_component method adds a new component object to the
        instrument and keeps track of its location within the
        sequence of components.  Normally a new component is added to
        the end of the sequence, but the before and after keywords can
        be used to select another location, here after is tested.
        """

        instr = setup_populated_instr()

        instr.add_component("test_component",
                            "test_for_reading",
                            after="first_component")

        self.assertEqual(len(instr.component_list), 4)
        self.assertEqual(instr.component_list[1].name, "test_component")
        self.assertEqual(instr.component_list[3].name, "third_component")

    def test_add_component_simple_after_error(self):
        """
        Checks add_component raises a NameError if after keyword specifies a
        non-existent component

        The add_component method adds a new component object to the
        instrument and keeps track of its location within the
        sequence of components.  Normally a new component is added to
        the end of the sequence, but the before and after keywords can
        be used to select another location, here before is tested.
        """

        instr = setup_populated_instr()

        with self.assertRaises(NameError):
            instr.add_component("test_component",
                                "test_for_reading",
                                after="non_existent_component")

    def test_add_component_simple_before_error(self):
        """
        Checks add_component raises a NameError if before keyword specifies a
        non-existent component

        The add_component method adds a new component object to the
        instrument and keeps track of its location within the
        sequence of components.  Normally a new component is added to
        the end of the sequence, but the before and after keywords can
        be used to select another location, here after is tested.
        """

        instr = setup_populated_instr()

        with self.assertRaises(NameError):
            instr.add_component("test_component",
                                "test_for_reading",
                                before="non_existent_component")

    def test_add_component_simple_double_naming_error(self):
        """
        This tests checks that an error occurs when giving a new
        component a name which has already been used.
        """

        instr = setup_populated_instr()

        with self.assertRaises(NameError):
            instr.add_component("first_component", "test_for_reading")

    def test_copy_component_simple(self):
        """
        Checks that a component can be copied using the name
        """

        instr = setup_populated_with_some_options_instr()

        comp = instr.copy_component("copy_of_second_comp", "second_component")

        self.assertEqual(comp.name, "copy_of_second_comp")
        self.assertEqual(comp.yheight, 1.23)
        self.assertEqual(comp.AT_data[0], 0)
        self.assertEqual(comp.AT_data[1], 0)
        self.assertEqual(comp.AT_data[2], 2)

    def test_infer_copy_component_simple(self):
        """
        Checks that a component can be copied using the name
        while giving the new instance name as the python variable
        """

        instr = setup_populated_with_some_options_instr()

        copy_of_second_comp = instr.copy_component("second_component")

        self.assertEqual(copy_of_second_comp.name, "copy_of_second_comp")
        self.assertEqual(copy_of_second_comp.yheight, 1.23)
        self.assertEqual(copy_of_second_comp.AT_data[0], 0)
        self.assertEqual(copy_of_second_comp.AT_data[1], 0)
        self.assertEqual(copy_of_second_comp.AT_data[2], 2)

    def test_copy_component_simple_fail(self):
        """
        Checks a NameError is raised if trying to copy a component that does
        not exist
        """

        instr = setup_populated_with_some_options_instr()

        with self.assertRaises(NameError):
            instr.copy_component("copy_of_second_comp", "unknown_component")

    def test_copy_component_simple_object(self):
        """
        Checks that a component can be copied using the object
        """

        instr = setup_populated_with_some_options_instr()

        comp = instr.get_component("second_component")

        comp = instr.copy_component("copy_of_second_comp", comp)

        self.assertEqual(comp.name, "copy_of_second_comp")
        self.assertEqual(comp.yheight, 1.23)
        self.assertEqual(comp.AT_data[0], 0)
        self.assertEqual(comp.AT_data[1], 0)
        self.assertEqual(comp.AT_data[2], 2)

    def test_copy_component_keywords(self):
        """
        Checks that a component can be copied and that keyword
        arguments given under copy operation is successfully
        applied to the new component. A check is also made to
        ensure that the original component was not modified.
        """

        instr = setup_populated_with_some_options_instr()

        comp = instr.copy_component("copy_of_second_comp", "second_component",
                                    AT=[1, 2, 3], SPLIT=10)

        self.assertEqual(comp.name, "copy_of_second_comp")
        self.assertEqual(comp.yheight, 1.23)
        self.assertEqual(comp.AT_data[0], 1)
        self.assertEqual(comp.AT_data[1], 2)
        self.assertEqual(comp.AT_data[2], 3)
        self.assertEqual(comp.SPLIT, 10)

        # ensure original component was not changed
        original = instr.get_component("second_component")
        self.assertEqual(original.name, "second_component")
        self.assertEqual(original.yheight, 1.23)
        self.assertEqual(original.AT_data[0], 0)
        self.assertEqual(original.AT_data[1], 0)
        self.assertEqual(original.AT_data[2], 2)
        self.assertEqual(original.SPLIT, 0)

    def test_remove_component(self):
        """
        Ensure a component can be removed
        """
        instr = setup_populated_instr()

        instr.remove_component("second_component")

        self.assertEqual(len(instr.component_list), 2)
        self.assertEqual(instr.component_list[0].name, "first_component")
        self.assertEqual(instr.component_list[1].name, "third_component")

    def test_move_component(self):
        """
        Ensure a component can be moved
        """
        instr = setup_populated_instr()

        instr.move_component("second_component", before="first_component")

        self.assertEqual(len(instr.component_list), 3)
        self.assertEqual(instr.component_list[0].name, "second_component")
        self.assertEqual(instr.component_list[1].name, "first_component")
        self.assertEqual(instr.component_list[2].name, "third_component")

    def test_RELATIVE_error(self):
        """
        Ensure check_for_errors finds impossible relative statement
        """

        instr = setup_populated_instr()

        second_component = instr.get_component("second_component")
        second_component.set_AT([0, 0, 0], RELATIVE="third_component")

        self.assertTrue(instr.has_errors())

        with self.assertRaises(McStasError):
            instr.check_for_errors()

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_RELATIVE_error_and_checks_false(self, mock_f):
        """
        Ensure check_for_errors finds impossible relative statement
        """

        instr = setup_populated_instr()

        second_component = instr.get_component("second_component")
        second_component.set_AT([0, 0, 0], RELATIVE="third_component")

        self.assertTrue(instr.has_errors())

        with self.assertRaises(McStasError):
            instr.write_full_instrument()

        instr.settings(checks=False)
        instr.write_full_instrument()

    def test_get_component_simple(self):
        """
        get_component retrieves a component with a given name for
        easier manipulation. Check it works as intended.
        """

        instr = setup_populated_instr()

        comp = instr.get_component("second_component")

        self.assertEqual(comp.name, "second_component")

    def test_get_component_simple_error(self):
        """
        get_component retrieves a component with a given name for
        easier manipulation. Check it fails when the component name
        doesn't correspond to a component in the instrument.
        """

        instr = setup_populated_instr()

        with self.assertRaises(NameError):
            instr.get_component("non_existing_component")

    def test_get_last_component_simple(self):
        """
        Check get_last_component retrieves the last component
        """

        instr = setup_populated_instr()

        comp = instr.get_last_component()

        self.assertEqual(comp.name, "third_component")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_print_component(self, mock_stdout):
        """
        print_component calls the print_long method in the component
        class.
        """

        instr = setup_populated_instr()
        instr.get_component("second_component").set_parameters(dist=5)

        instr.print_component("second_component")

        output = mock_stdout.getvalue().split("\n")

        self.assertEqual(output[0],
                         "COMPONENT second_component = test_for_reading(")

        par_name = bcolors.BOLD + "dist" + bcolors.ENDC
        value = (bcolors.BOLD + bcolors.OKGREEN
                 + "5" + bcolors.ENDC + bcolors.ENDC)
        self.assertEqual(output[1], "  " + par_name + " = " + value + " // [m]")

        par_name = bcolors.BOLD + "gauss" + bcolors.ENDC
        warning = (bcolors.FAIL
                   + " : Required parameter not yet specified"
                   + bcolors.ENDC)
        self.assertEqual(output[2], "  " + par_name + warning)

        par_name = bcolors.BOLD + "test_string" + bcolors.ENDC
        warning = (bcolors.FAIL
                   + " : Required parameter not yet specified"
                   + bcolors.ENDC)
        self.assertEqual(output[3], "  " + par_name + warning)

        self.assertEqual(output[4], ")")
        self.assertEqual(output[5], "AT (0, 0, 0) ABSOLUTE")

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_print_component_short(self, mock_stdout):
        """
        print_component_short calls the print_short method in the
        component class.
        """

        instr = setup_populated_instr()
        instr.get_component("second_component").set_AT([-1, 2, 3.4],
                                                       RELATIVE="home")

        instr.print_component_short("second_component")

        output = mock_stdout.getvalue().split("\n")

        expected = ("second_component = test_for_reading "
                    + "\tAT [-1, 2, 3.4] RELATIVE home "
                    + "ROTATED [0, 0, 0] RELATIVE home")

        self.assertEqual(output[0], expected)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_show_components_simple(self, mock_stdout):
        """
        Tests show_components for simple case

        show_components calls the print_short method in the component
        class for each component and aligns the data for display
        """

        instr = setup_populated_instr()

        instr.show_components(line_length=300)

        output = mock_stdout.getvalue().split("\n")

        expected = ("first_component  test_for_reading"
                    + " AT (0, 0, 0) ABSOLUTE")
        self.assertEqual(output[0], expected)

        expected = ("second_component test_for_reading"
                    + " AT (0, 0, 0) ABSOLUTE")
        self.assertEqual(output[1], expected)

        expected = ("third_component  test_for_reading"
                    + " AT (0, 0, 0) ABSOLUTE")
        self.assertEqual(output[2], expected)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_show_components_complex(self, mock_stdout):
        """
        Tests show_components for complex case

        show_components calls the print_short method in the component
        class for each component and aligns the data for display
        """

        instr = setup_populated_instr()

        instr.get_component("first_component").set_AT([-0.1, 12, "dist"],
                                                      RELATIVE="home")
        instr.get_component("second_component").set_ROTATED([-4, 0.001, "theta"],
                                                            RELATIVE="etc")
        comp = instr.get_last_component()
        comp.component_name = "test_name"

        instr.show_components(line_length=300)

        output = mock_stdout.getvalue().split("\n")

        expected = ("first_component  test_for_reading"
                    + " AT (-0.1, 12, dist) RELATIVE home")
        self.assertEqual(output[0], expected)

        expected = ("second_component test_for_reading"
                    + " AT (0, 0, 0)        ABSOLUTE"
                    + "      ROTATED (-4, 0.001, theta) RELATIVE etc")
        self.assertEqual(output[1], expected)

        expected = ("third_component  test_name"
                    + "        AT (0, 0, 0)        ABSOLUTE     ")
        self.assertEqual(output[2], expected)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_show_components_complex_2lines(self, mock_stdout):
        """
        show_components calls the print_short method in the component
        class for each component and aligns the data for display

        This version of the tests forces two lines of output.
        """

        instr = setup_populated_instr()

        instr.get_component("first_component").set_AT([-0.1, 12, "dist"],
                                                      RELATIVE="home")
        instr.get_component("second_component").set_ROTATED([-4, 0.001, "theta"],
                                                            RELATIVE="etc")
        comp = instr.get_last_component()
        comp.component_name = "test_name"

        instr.show_components(line_length=80)

        output = mock_stdout.getvalue().split("\n")

        expected = ("first_component  test_for_reading"
                    + " AT      (-0.1, 12, dist)   RELATIVE home")
        self.assertEqual(output[0], expected)

        expected = ("second_component test_for_reading"
                    + " AT      (0, 0, 0)          ABSOLUTE      ")
        self.assertEqual(output[1], expected)

        expected = ("                                 "
                    + " ROTATED (-4, 0.001, theta) RELATIVE etc")
        self.assertEqual(output[2], expected)

        expected = ("third_component  test_name       "
                    + " AT      (0, 0, 0)          ABSOLUTE     ")
        self.assertEqual(output[3], expected)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_show_components_complex_3lines(self, mock_stdout):
        """
        show_components calls the print_short method in the component
        class for each component and aligns the data for display

        This version of the tests forces three lines of output.
        """

        instr = setup_populated_instr()

        instr.get_component("first_component").set_AT([-0.1, 12, "dist"],
                                                      RELATIVE="home")
        instr.get_component("second_component").set_ROTATED([-4, 0.001, "theta"],
                                                            RELATIVE="etc")
        comp = instr.get_last_component()
        comp.component_name = "test_name"

        instr.show_components(line_length=1)  # Three lines maximum

        output = mock_stdout.getvalue().split("\n")

        expected = (bcolors.BOLD
                    + "first_component"
                    + bcolors.ENDC
                    + "  "
                    + bcolors.BOLD
                    + "test_for_reading"
                    + bcolors.ENDC
                    + " ")
        self.assertEqual(output[0], expected)

        expected = "  AT      (-0.1, 12, dist) RELATIVE home"
        self.assertEqual(output[1], expected)

        expected = (bcolors.BOLD
                    + "second_component"
                    + bcolors.ENDC
                    + "  "
                    + bcolors.BOLD
                    + "test_for_reading"
                    + bcolors.ENDC
                    + " ")
        self.assertEqual(output[2], expected)

        expected = "  AT      (0, 0, 0) ABSOLUTE "
        self.assertEqual(output[3], expected)

        expected = "  ROTATED (-4, 0.001, theta) RELATIVE etc"
        self.assertEqual(output[4], expected)

        expected = (bcolors.BOLD
                    + "third_component"
                    + bcolors.ENDC
                    + "  "
                    + bcolors.BOLD
                    + "test_name"
                    + bcolors.ENDC
                    + " ")
        self.assertEqual(output[5], expected)

        expected = "  AT      (0, 0, 0) ABSOLUTE"
        self.assertEqual(output[6], expected)

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_write_c_files_simple(self, mock_f):
        """
        Write_c_files writes the strings for declare, initialize,
        and trace to files that are then included in McStas files.
        This is an obsolete method, but may be repurposed later
        so that instrument parts can be created with the modern
        syntax.

        The generated includes file in the test directory is written
        by this test. It will fail if it does not have rights to
        create the directory.
        """

        instr = setup_populated_instr()

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        current_directory = os.getcwd()
        os.chdir(THIS_DIR)

        try:
            instr.write_c_files()
        finally:
            os.chdir(current_directory)

        base_path = os.path.join(".", "generated_includes")
        expected_path = os.path.join(base_path, "test_instrument_declare.c")
        mock_f.assert_any_call(expected_path, "w")
        mock_f.assert_any_call(expected_path, "a")

        expected_path = os.path.join(base_path, "test_instrument_initialize.c")
        mock_f.assert_any_call(expected_path, "w")

        expected_path = os.path.join(base_path, "test_instrument_trace.c")
        mock_f.assert_any_call(expected_path, "w")

        expected_path = os.path.join(base_path, "test_instrument_component_trace.c")
        mock_f.assert_any_call(expected_path, "w")

        # This does not check that the right thing is written to the
        # right file. Can be improved by splitting the method into
        # several for easier testing. Acceptable since it is rarely
        # used.
        handle = mock_f()
        call = unittest.mock.call
        wrts = [
         call("// declare section for test_instrument \n"),
         call("double two_theta;"),
         call("\n"),
         call("// Start of initialize for generated test_instrument\n"
              + "two_theta = 2.0*theta;\n"),
         call("// Start of trace section for generated test_instrument\n"),
         call("COMPONENT first_component = test_for_reading("),
         call(")\n"),
         call("AT (0,0,0)"),
         call(" ABSOLUTE\n"),
         call("\n"),
         call("COMPONENT second_component = test_for_reading("),
         call(")\n"),
         call("AT (0,0,0)"),
         call(" ABSOLUTE\n"),
         call("\n"),
         call("COMPONENT third_component = test_for_reading("),
         call(")\n"),
         call("AT (0,0,0)"),
         call(" ABSOLUTE\n"),
         call("\n")]

        handle.write.assert_has_calls(wrts, any_order=False)

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    @unittest.mock.patch('datetime.datetime')
    def test_write_full_instrument_simple(self, mock_datetime, mock_f):
        """
        The write_full_instrument method write the information
        contained in the instrument instance to a file with McStas
        syntax.

        The test includes a time stamp in the written and expected
        data that has an accuracy of 1 second.  It is unlikely to fail
        due to this, but it can happen.
        """

        # Fix datetime for call
        fixed_datetime = datetime.datetime(2023, 12, 14, 12, 44, 21)
        mock_datetime.now.return_value = fixed_datetime

        instr = setup_populated_instr()
        instr.write_full_instrument()

        t_format = "%H:%M:%S on %B %d, %Y"

        my_call = unittest.mock.call
        wrts = [
         my_call("/" + 80*"*" + "\n"),
         my_call("* \n"),
         my_call("* McStas, neutron ray-tracing package\n"),
         my_call("*         Copyright (C) 1997-2008, All rights reserved\n"),
         my_call("*         Risoe National Laboratory, Roskilde, Denmark\n"),
         my_call("*         Institut Laue Langevin, Grenoble, France\n"),
         my_call("* \n"),
         my_call("* This file was written by McStasScript, which is a \n"),
         my_call("* python based McStas instrument generator written by \n"),
         my_call("* Mads Bertelsen in 2019 while employed at the \n"),
         my_call("* European Spallation Source Data Management and \n"),
         my_call("* Software Centre\n"),
         my_call("* \n"),
         my_call("* Instrument test_instrument\n"),
         my_call("* \n"),
         my_call("* %Identification\n"),
         my_call("* Written by: Python McStas Instrument Generator\n"),
         my_call("* Date: %s\n" % fixed_datetime.strftime(t_format)),
         my_call("* Origin: ESS DMSC\n"),
         my_call("* %INSTRUMENT_SITE: Generated_instruments\n"),
         my_call("* \n"),
         my_call("* \n"),
         my_call("* %Parameters\n"),
         my_call("* \n"),
         my_call("* %End \n"),
         my_call("*"*80 + "/\n"),
         my_call("\n"),
         my_call("DEFINE INSTRUMENT test_instrument ("),
         my_call("\n"),
         my_call("double theta"),
         my_call(", "),
         my_call(""),
         my_call("\n"),
         my_call("double has_default"),
         my_call(" = 37"),
         my_call(" "),
         my_call(""),
         my_call("\n"),
         my_call(")\n"),
         my_call("\n"),
         my_call("DECLARE \n%{\n"),
         my_call("double two_theta;"),
         my_call("\n"),
         my_call("%}\n\n"),
         my_call("INITIALIZE \n%{\n"),
         my_call("// Start of initialize for generated test_instrument\n"
                 + "two_theta = 2.0*theta;\n"),
         my_call("%}\n\n"),
         my_call("TRACE \n"),
         my_call("COMPONENT first_component = test_for_reading("),
         my_call(")\n"),
         my_call("AT (0,0,0)"),
         my_call(" ABSOLUTE\n"),
         my_call("\n"),
         my_call("COMPONENT second_component = test_for_reading("),
         my_call(")\n"),
         my_call("AT (0,0,0)"),
         my_call(" ABSOLUTE\n"),
         my_call("\n"),
         my_call("COMPONENT third_component = test_for_reading("),
         my_call(")\n"),
         my_call("AT (0,0,0)"),
         my_call(" ABSOLUTE\n"),
         my_call("\n"),
         my_call("FINALLY \n%{\n"),
         my_call("// Start of finally for generated test_instrument\n"),
         my_call("%}\n"),
         my_call("\nEND\n")]

        expected_path = os.path.join(".", "test_instrument.instr")
        mock_f.assert_called_with(expected_path, "w")
        handle = mock_f()
        handle.write.assert_has_calls(wrts, any_order=False)

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_write_full_instrument_dependency(self, mock_f):
        """
        The write_full_instrument method write the information
        contained in the instrument instance to a file with McStas
        syntax. Here tested with the dependency section enabled.

        The test includes a time stamp in the written and expected
        data that has an accuracy of 1 second.  It is unlikely to fail
        due to this, but it can happen.
        """

        instr = setup_populated_instr()
        instr.set_dependency("-DMCPLPATH=GETPATH(data)")
        instr.write_full_instrument()

        t_format = "%H:%M:%S on %B %d, %Y"

        my_call = unittest.mock.call
        wrts = [
            my_call("/" + 80 * "*" + "\n"),
            my_call("* \n"),
            my_call("* McStas, neutron ray-tracing package\n"),
            my_call("*         Copyright (C) 1997-2008, All rights reserved\n"),
            my_call("*         Risoe National Laboratory, Roskilde, Denmark\n"),
            my_call("*         Institut Laue Langevin, Grenoble, France\n"),
            my_call("* \n"),
            my_call("* This file was written by McStasScript, which is a \n"),
            my_call("* python based McStas instrument generator written by \n"),
            my_call("* Mads Bertelsen in 2019 while employed at the \n"),
            my_call("* European Spallation Source Data Management and \n"),
            my_call("* Software Centre\n"),
            my_call("* \n"),
            my_call("* Instrument test_instrument\n"),
            my_call("* \n"),
            my_call("* %Identification\n"),
            my_call("* Written by: Python McStas Instrument Generator\n"),
            my_call("* Date: %s\n" % datetime.datetime.now().strftime(t_format)),
            my_call("* Origin: ESS DMSC\n"),
            my_call("* %INSTRUMENT_SITE: Generated_instruments\n"),
            my_call("* \n"),
            my_call("* \n"),
            my_call("* %Parameters\n"),
            my_call("* \n"),
            my_call("* %End \n"),
            my_call("*" * 80 + "/\n"),
            my_call("\n"),
            my_call("DEFINE INSTRUMENT test_instrument ("),
            my_call("\n"),
            my_call("double theta"),
            my_call(", "),
            my_call(""),
            my_call("\n"),
            my_call("double has_default"),
            my_call(" = 37"),
            my_call(" "),
            my_call(""),
            my_call("\n"),
            my_call(")\n"),
            my_call('DEPENDENCY "-DMCPLPATH=GETPATH(data)"\n'),
            my_call("\n"),
            my_call("DECLARE \n%{\n"),
            my_call("double two_theta;"),
            my_call("\n"),
            my_call("%}\n\n"),
            my_call("INITIALIZE \n%{\n"),
            my_call("// Start of initialize for generated test_instrument\n"
                    + "two_theta = 2.0*theta;\n"),
            my_call("%}\n\n"),
            my_call("TRACE \n"),
            my_call("COMPONENT first_component = test_for_reading("),
            my_call(")\n"),
            my_call("AT (0,0,0)"),
            my_call(" ABSOLUTE\n"),
            my_call("\n"),
            my_call("COMPONENT second_component = test_for_reading("),
            my_call(")\n"),
            my_call("AT (0,0,0)"),
            my_call(" ABSOLUTE\n"),
            my_call("\n"),
            my_call("COMPONENT third_component = test_for_reading("),
            my_call(")\n"),
            my_call("AT (0,0,0)"),
            my_call(" ABSOLUTE\n"),
            my_call("\n"),
            my_call("FINALLY \n%{\n"),
            my_call("// Start of finally for generated test_instrument\n"),
            my_call("%}\n"),
            my_call("\nEND\n")]

        expected_path = os.path.join(".", "test_instrument.instr")
        mock_f.assert_called_with(expected_path, "w")
        handle = mock_f()
        handle.write.assert_has_calls(wrts, any_order=False)

    @unittest.mock.patch('__main__.__builtins__.open',
                         new_callable=unittest.mock.mock_open)
    def test_write_full_instrument_search(self, mock_f):
        """
        The write_full_instrument method write the information
        contained in the instrument instance to a file with McStas
        syntax. Here tested with the search section enabled.

        The test includes a time stamp in the written and expected
        data that has an accuracy of 1 second.  It is unlikely to fail
        due to this, but it can happen.
        """

        instr = setup_populated_instr()
        instr.add_search("first_search")
        instr.add_search("second search", SHELL=True)
        instr.write_full_instrument()

        t_format = "%H:%M:%S on %B %d, %Y"

        my_call = unittest.mock.call
        wrts = [
            my_call("/" + 80 * "*" + "\n"),
            my_call("* \n"),
            my_call("* McStas, neutron ray-tracing package\n"),
            my_call("*         Copyright (C) 1997-2008, All rights reserved\n"),
            my_call("*         Risoe National Laboratory, Roskilde, Denmark\n"),
            my_call("*         Institut Laue Langevin, Grenoble, France\n"),
            my_call("* \n"),
            my_call("* This file was written by McStasScript, which is a \n"),
            my_call("* python based McStas instrument generator written by \n"),
            my_call("* Mads Bertelsen in 2019 while employed at the \n"),
            my_call("* European Spallation Source Data Management and \n"),
            my_call("* Software Centre\n"),
            my_call("* \n"),
            my_call("* Instrument test_instrument\n"),
            my_call("* \n"),
            my_call("* %Identification\n"),
            my_call("* Written by: Python McStas Instrument Generator\n"),
            my_call("* Date: %s\n" % datetime.datetime.now().strftime(t_format)),
            my_call("* Origin: ESS DMSC\n"),
            my_call("* %INSTRUMENT_SITE: Generated_instruments\n"),
            my_call("* \n"),
            my_call("* \n"),
            my_call("* %Parameters\n"),
            my_call("* \n"),
            my_call("* %End \n"),
            my_call("*" * 80 + "/\n"),
            my_call("\n"),
            my_call("DEFINE INSTRUMENT test_instrument ("),
            my_call("\n"),
            my_call("double theta"),
            my_call(", "),
            my_call(""),
            my_call("\n"),
            my_call("double has_default"),
            my_call(" = 37"),
            my_call(" "),
            my_call(""),
            my_call("\n"),
            my_call(")\n"),
            my_call("\n"),
            my_call("DECLARE \n%{\n"),
            my_call("double two_theta;"),
            my_call("\n"),
            my_call("%}\n\n"),
            my_call("INITIALIZE \n%{\n"),
            my_call("// Start of initialize for generated test_instrument\n"
                    + "two_theta = 2.0*theta;\n"),
            my_call("%}\n\n"),
            my_call("TRACE \n"),
            my_call('SEARCH "first_search"\n'),
            my_call('SEARCH SHELL "second search"\n'),
            my_call("COMPONENT first_component = test_for_reading("),
            my_call(")\n"),
            my_call("AT (0,0,0)"),
            my_call(" ABSOLUTE\n"),
            my_call("\n"),
            my_call("COMPONENT second_component = test_for_reading("),
            my_call(")\n"),
            my_call("AT (0,0,0)"),
            my_call(" ABSOLUTE\n"),
            my_call("\n"),
            my_call("COMPONENT third_component = test_for_reading("),
            my_call(")\n"),
            my_call("AT (0,0,0)"),
            my_call(" ABSOLUTE\n"),
            my_call("\n"),
            my_call("FINALLY \n%{\n"),
            my_call("// Start of finally for generated test_instrument\n"),
            my_call("%}\n"),
            my_call("\nEND\n")]

        expected_path = os.path.join(".", "test_instrument.instr")
        mock_f.assert_called_with(expected_path, "w")
        handle = mock_f()
        handle.write.assert_has_calls(wrts, any_order=False)

    # mock sys.stdout to avoid printing to terminal
    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_run_full_instrument_required_par_error(self, mock_stdout):
        """
        Tests run_full_instrument raises error when lacking required parameter

        The populated instr has a required parameter, and when not
        given it should raise an error.
        """
        instr = setup_populated_instr()

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with self.assertRaises(NameError):
            instr.run_full_instrument(output_path="test_data_set",
                                      increment_folder_name=False,
                                      executable_path=executable_path)

    def test_run_full_instrument_junk_par_error(self):
        """
        Check run_full_instrument raises a NameError if a unrecognized
        parameter is provided, here junk.
        """
        instr = setup_populated_instr()

        pars = {"theta": 2, "junk": "test"}

        with self.assertRaises(KeyError):
            instr.run_full_instrument(output_path="test_data_set",
                                      increment_folder_name=False,
                                      parameters=pars)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    @unittest.mock.patch("subprocess.run")
    def test_x_ray_run_full_instrument_basic(self, mock_sub, mock_stdout):
        """
        Tests x-ray run_full_instrument

        Check a simple run performs the correct system call.  Here
        the output_path is set to a name that does not correspond to a
        existing file so no error is thrown.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            new_folder_name = "folder_name_which_is_unused"
            if os.path.exists(new_folder_name):
                raise RuntimeError("Folder_name was supposed to not "
                                   + "exist before "
                                   + "test_run_backengine_basic")

            instr = setup_populated_x_ray_instr_with_dummy_path()
            instr.run_full_instrument(output_path=new_folder_name,
                                      increment_folder_name=False,
                                      executable_path=executable_path,
                                      parameters={"theta": 1})

        expected_path = os.path.join(executable_path, "mxrun")
        expected_path = '"' + expected_path + '"'

        expected_folder_path = os.path.join(THIS_DIR, new_folder_name)

        # a double space because of a missing option
        expected_call = (expected_path + " -c -n 1000000 "
                         + "-d " + expected_folder_path
                         + "  test_instrument.instr"
                         + " theta=1 has_default=37")

        expected_run_path = os.path.join(THIS_DIR, ".")

        mock_sub.assert_called_with(expected_call,
                                    shell=True,
                                    cwd=expected_run_path,
                                    stderr=-2, stdout=-1,
                                    universal_newlines=True)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_run_backengine_existing_folder(self, mock_stdout):
        """
        Test neutron run of backengine fails if using existing folder
        for output_path and with increment_folder_name disabled.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            instr = setup_populated_instr_with_dummy_path()

            instr.set_parameters({"theta": 1})
            instr.settings(output_path="test_data_set",
                           increment_folder_name=False,
                           executable_path=executable_path)

            with self.assertRaises(NameError):
                instr.backengine()

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    @unittest.mock.patch("subprocess.run")
    def test_run_backengine_basic(self, mock_sub, mock_stdout):
        """
        Test neutron run_full_instrument

        Check a simple run performs the correct system call. Here
        the output_path is set to a name that does not correspond to a
        existing file so no error is thrown.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            new_folder_name = "folder_name_which_is_unused"
            if os.path.exists(new_folder_name):
                raise RuntimeError("Folder_name was supposed to not "
                                   + "exist before "
                                   + "test_run_backengine_basic")

            instr = setup_populated_instr_with_dummy_path()

            instr.set_parameters({"theta": 1})
            instr.settings(output_path=new_folder_name,
                           increment_folder_name=True,
                           executable_path=executable_path)
            instr.backengine()

        expected_path = os.path.join(executable_path, "mcrun")
        expected_path = '"' + expected_path + '"'

        expected_folder_path = os.path.join(THIS_DIR, new_folder_name)

        # a double space because of a missing option
        expected_call = (expected_path + " -c -n 1000000 "
                         + "-d " + expected_folder_path
                         + "  test_instrument.instr"
                         + " theta=1 has_default=37")

        mock_sub.assert_called_with(expected_call,
                                    shell=True,
                                    stderr=-2, stdout=-1,
                                    universal_newlines=True,
                                    cwd=run_path)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    @unittest.mock.patch("subprocess.run")
    def test_run_backengine_complex_settings(self, mock_sub, mock_stdout):
        """
        Test settings are passed to backengine with complex settings

        Check run performs the correct system call with settings. Here
        the output_path is set to a name that does not correspond to a
        existing file so no error is thrown.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            new_folder_name = "folder_name_which_is_unused"
            if os.path.exists(new_folder_name):
                raise RuntimeError("Folder_name was supposed to not "
                                   + "exist before "
                                   + "test_run_backengine_basic")

            instr = setup_populated_instr_with_dummy_path()

            instr.set_parameters({"theta": 1})
            instr.settings(output_path=new_folder_name,
                           increment_folder_name=True,
                           executable_path=executable_path,
                           mpi=5, ncount=19373, openacc=True,
                           NeXus=True, force_compile=False,
                           seed=300, gravity=True, checks=False,
                           save_comp_pars=True)
            instr.backengine()

        expected_path = os.path.join(executable_path, "mcrun")
        expected_path = '"' + expected_path + '"'

        expected_folder_path = os.path.join(THIS_DIR, new_folder_name)

        # a double space because of a missing option
        expected_call = (expected_path + " -g --format=NeXus "
                         + "--openacc "
                         + "-n 19373 --mpi=5 --seed=300 "
                         + "-d " + expected_folder_path
                         + "  test_instrument.instr"
                         + " theta=1 has_default=37")

        mock_sub.assert_called_with(expected_call,
                                    shell=True,
                                    stderr=-2, stdout=-1,
                                    universal_newlines=True,
                                    cwd=run_path)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    @unittest.mock.patch("subprocess.run")
    def test_run_full_instrument_complex(self, mock_sub, mock_stdout):
        """
        Test neutron run_full_instrument in more complex case

        Check a complex run performs the correct system call. Here
        the output_path is set to a name that does not correspond to a
        existing file so no error is thrown.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            new_folder_name = "folder_name_which_is_unused"
            if os.path.exists(new_folder_name):
                raise RuntimeError("Folder_name was supposed to not "
                                   + "exist before "
                                   + "test_run_backengine_basic")

            instr = setup_populated_instr_with_dummy_path()

            # Add some extra parameters for testing
            instr.add_parameter("A")
            instr.add_parameter("BC")

            instr.run_full_instrument(output_path=new_folder_name,
                                      increment_folder_name=False,
                                      executable_path=executable_path,
                                      mpi=7,
                                      seed=300,
                                      ncount=48.4,
                                      gravity=True,
                                      custom_flags="-fo",
                                      parameters={"A": 2,
                                                  "BC": "car",
                                                  "theta": "\"toy\""})

        expected_path = os.path.join(executable_path, "mcrun")
        expected_path = '"' + expected_path + '"'
        expected_folder_path = os.path.join(THIS_DIR, new_folder_name)

        # a double space because of a missing option
        expected_call = (expected_path + " -c -g -n 48 --mpi=7 --seed=300 "
                         + "-d " + expected_folder_path
                         + " -fo test_instrument.instr "
                         + "theta=\"toy\" has_default=37 A=2 BC=car")

        mock_sub.assert_called_with(expected_call,
                                    shell=True,
                                    stderr=-2, stdout=-1,
                                    universal_newlines=True,
                                    cwd=run_path)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    @unittest.mock.patch("subprocess.run")
    def test_run_full_instrument_overwrite_default(self, mock_sub,
                                                   mock_stdout):
        """
        Check that default parameters are overwritten by given
        parameters in run_full_instrument.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            new_folder_name = "folder_name_which_is_unused"
            if os.path.exists(new_folder_name):
                raise RuntimeError("Folder_name was supposed to not "
                                   + "exist before "
                                   + "test_run_backengine_basic")

            instr = setup_populated_instr_with_dummy_path()

            # Add some extra parameters for testing
            instr.add_parameter("A")
            instr.add_parameter("BC")

            instr.run_full_instrument(output_path=new_folder_name,
                                      increment_folder_name=False,
                                      executable_path=executable_path,
                                      mpi=7,
                                      ncount=48.4,
                                      custom_flags="-fo",
                                      parameters={"A": 2,
                                                  "BC": "car",
                                                  "theta": "\"toy\"",
                                                  "has_default": 10})

        expected_path = os.path.join(executable_path, "mcrun")
        expected_path = '"' + expected_path + '"'
        expected_folder_path = os.path.join(THIS_DIR, new_folder_name)

        # a double space because of a missing option
        expected_call = (expected_path + " -c -n 48 --mpi=7 "
                         + "-d " + expected_folder_path
                         + " -fo test_instrument.instr "
                         + "theta=\"toy\" has_default=10 A=2 BC=car")

        mock_sub.assert_called_with(expected_call,
                                    shell=True,
                                    stderr=-2, stdout=-1,
                                    universal_newlines=True,
                                    cwd=run_path)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    @unittest.mock.patch("subprocess.run")
    def test_run_full_instrument_x_ray_basic(self, mock_sub, mock_stdout):
        """
        Test x-ray run_full_instrument

        Check a simple run performs the correct system call.  Here
        the output_path is set to a name that does not correspond to a
        existing file so no error is thrown.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            new_folder_name = "folder_name_which_is_unused"
            if os.path.exists(new_folder_name):
                raise RuntimeError("Folder_name was supposed to not "
                                   + "exist before "
                                   + "test_run_backengine_basic")

            instr = setup_populated_x_ray_instr_with_dummy_path()

            instr.run_full_instrument(output_path=new_folder_name,
                                      increment_folder_name=False,
                                      executable_path=executable_path,
                                      parameters={"theta": 1})

        expected_path = os.path.join(executable_path, "mxrun")
        expected_path = '"' + expected_path + '"'
        expected_folder_path = os.path.join(THIS_DIR, new_folder_name)

        # a double space because of a missing option
        expected_call = (expected_path + " -c -n 1000000 "
                         + "-d " + expected_folder_path
                         + "  test_instrument.instr"
                         + " theta=1 has_default=37")

        mock_sub.assert_called_with(expected_call,
                                    shell=True,
                                    stderr=-2, stdout=-1,
                                    universal_newlines=True,
                                    cwd=run_path)

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    @unittest.mock.patch("subprocess.run")
    def test_show_instrument_basic(self, mock_sub, mock_stdout):
        """
        Test show_instrument methods makes correct system calls
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        instr = setup_populated_instr_with_dummy_path()

        instr.set_parameters(theta=1.2)
        instr.show_instrument()

        os.chdir(current_work_dir)

        expected_path = os.path.join(executable_path, "bin", "mcdisplay-webgl")
        expected_path = '"' + expected_path + '"'
        expected_instr_path = os.path.join(THIS_DIR, "test_instrument.instr")

        # a double space because of a missing option
        expected_call = (expected_path
                         + " --dirname test_instrument_mcdisplay"
                         + " " + expected_instr_path
                         + "  theta=1.2 has_default=37")

        mock_sub.assert_called_with(expected_call,
                                    shell=True,
                                    stderr=-1, stdout=-1,
                                    universal_newlines=True,
                                    cwd=".")

    def test_show_dumps_works(self):
        """
        Ensures show_dumps runs
        """

        instr = setup_populated_instr_with_dummy_MCPL_comps()
        insert_mock_dump(instr, "second_component")

        instr.show_dumps()

    def test_show_dump_works(self):
        """
        Ensures show_dump runs and that db can get the stored dump
        """

        instr = setup_populated_instr_with_dummy_MCPL_comps()
        insert_mock_dump(instr, "second_component", run_name="custom_run", tag=31)

        instr.show_dump("second_component", "custom_run", 31)

        dump = instr.dump_database.get_dump("second_component", "custom_run", 31)
        self.assertEqual(dump.data["dump_point"], "second_component")
        self.assertEqual(dump.data["run_name"], "custom_run")
        self.assertEqual(dump.data["tag"], 31)

    def test_set_run_to_component(self):
        """
        Testing run_to method updates instr state correctly
        """

        instr = setup_populated_instr_with_dummy_MCPL_comps()

        second_component = instr.get_component("second_component")

        # Ensure remaining instrument can work
        third_component = instr.get_component("third_component")
        third_component.set_RELATIVE("second_component")

        instr.run_to(second_component)

        self.assertEqual(instr.run_from_ref, None)
        self.assertEqual(instr.run_to_ref, "second_component")
        self.assertEqual(instr.run_to_name, "Run")

        # Check filename added to parameters
        self.assertTrue("run_to_mcpl" in instr.parameters.parameters)
        self.assertEqual(instr.parameters.parameters["run_to_mcpl"].type, "string")

        instr.run_to("third_component", run_name="Test_name")
        self.assertEqual(instr.run_to_ref, "third_component")
        self.assertEqual(instr.run_to_name, "Test_name")

    def test_set_run_to_component_keywords(self):
        """
        Testing run_to method updates instr state with passed keywords
        """

        instr = setup_populated_instr_with_dummy_MCPL_comps()

        second_component = instr.get_component("second_component")

        # Ensure remaining instrument can work
        third_component = instr.get_component("third_component")
        third_component.set_RELATIVE("second_component")

        instr.run_to(second_component, test_keyword=58)

        self.assertIn("test_keyword", instr.run_to_component_parameters)
        self.assertEqual(instr.run_to_component_parameters["test_keyword"], 58)

    def test_set_run_to_nonexistant_component_fails(self):
        """
        Ensures run_to fails if the component doesn't exist
        """

        instr = setup_populated_instr_with_dummy_MCPL_comps()

        with self.assertRaises(ValueError):
            instr.run_to("component_that_does_not_exists")

    def test_set_run_to_component_with_ABSOLUTE_in_remaining_fails(self):
        """
        Ensures run_to fails if the remaining instrument refers to absolute
        """
        instr = setup_populated_instr_with_dummy_MCPL_comps()

        with self.assertRaises(McStasError):
            # Fails because the rest of the instrument has reference to ABSOLUTE
            instr.run_to("second_component")

    def test_set_run_to_component_with_early_ref_in_remaining_fails(self):
        """
        Ensures run_to fails if the remaining instrument refers to absolute
        """
        instr = setup_populated_instr_with_dummy_MCPL_comps()

        comp = instr.get_component("third_component")
        comp.set_RELATIVE("first_component")

        with self.assertRaises(McStasError):
            # Fails because the rest of the instrument has reference to component before split
            instr.run_to("second_component")

    def test_set_run_from_component(self):
        """
        Testing run_from method updates instr state correctly
        """
        instr = setup_populated_instr_with_dummy_MCPL_comps()

        insert_mock_dump(instr, "second_component")

        second_component = instr.get_component("second_component")

        # Ensure remaining instrument can work
        third_component = instr.get_component("third_component")
        third_component.set_RELATIVE("second_component")

        instr.run_from(second_component)

        self.assertEqual(instr.run_to_ref, None)
        self.assertEqual(instr.run_from_ref, "second_component")

        # Check filename added to parameters
        self.assertTrue("run_from_mcpl" in instr.parameters.parameters)
        self.assertEqual(instr.parameters.parameters["run_from_mcpl"].type, "string")

    def test_set_run_from_component_fails_if_no_dump(self):
        """
        Ensure run_from fails if there are no dumps at that location
        """
        instr = setup_populated_instr_with_dummy_MCPL_comps()

        with self.assertRaises(KeyError):
            instr.run_from("second_component")

        # Also fails if database is populated at a different component
        insert_mock_dump(instr, "second_component")

        with self.assertRaises(KeyError):
            instr.run_from("first_component")

    def test_set_run_from_component_keywords(self):
        """
        Testing run_to method updates instr state with passed keywords
        """

        instr = setup_populated_instr_with_dummy_MCPL_comps()

        insert_mock_dump(instr, "second_component")

        second_component = instr.get_component("second_component")

        # Ensure remaining instrument can work
        third_component = instr.get_component("third_component")
        third_component.set_RELATIVE("second_component")

        instr.run_from(second_component, test_keyword=37)

        self.assertIn("test_keyword", instr.run_from_component_parameters)
        self.assertEqual(instr.run_from_component_parameters["test_keyword"], 37)

    def test_set_run_from_and_run_to(self):
        """
        Ensure it is possible to use run_from and run_to, and that reset work
        """

        instr = setup_populated_instr_with_dummy_MCPL_comps()
        insert_mock_dump(instr, "second_component")
        third_component = instr.get_component("third_component")
        third_component.set_RELATIVE("second_component")

        instr.run_from("second_component")
        instr.run_to("third_component")

        self.assertEqual(instr.run_from_ref, "second_component")
        self.assertEqual(instr.run_to_ref, "third_component")

        instr.reset_run_points()

        self.assertEqual(instr.run_from_ref, None)
        self.assertEqual(instr.run_to_ref, None)

if __name__ == '__main__':
    unittest.main()



import matplotlib.pyplot as plt
from mcstasscript.helper.mcstas_objects import Component

class ComponentBox:
    """
    Helper class for creating text boxes describing components
    """
    def __init__(self, box_input):
        """
        Text box object
        """
        # set defaults
        self.position_x = None
        self.position_y = None
        self.box_height = None
        self.box_indent = None
        self.background_color = "white"
        self.outline_style = "-"
        self.outline_width = 1

        self.t = None  # text object
        self.graph_box_end = None  # graph position where text ends
        self.graph_box_start = None  # graph position where text ends

        # Load component
        if isinstance(box_input, str):
            self.component_object = None
            self.name = box_input
        elif isinstance(box_input, Component):
            self.component_object = box_input
            self.name = self.component_object.name

            # Decorate the box depending on the McStas features used
            if self.component_object.WHEN != "":
                self.outline_style = "--"

            if self.component_object.EXTEND != "":
                self.outline_width = 2.5
        else:
            raise ValueError("Input for box needs to be of type Component or str, not "
                             + str(type(box_input)))

        # Produced weighted length of name, capital letters count for 1.2
        self.weighted_name_length = len(self.name) + 0.2*sum(1 for c in self.name if c.isupper())

    def set_box_height(self, box_height):
        self.box_height = box_height

    def set_x(self, x):
        self.position_x = x

    def set_y(self, y):
        self.position_y = y

    def set_box_indent(self, value):
        self.box_indent = value

    def set_background_color(self, value):
        self.background_color = value

    def plot_box(self, ax):
        bbox = dict(boxstyle="round", facecolor=self.background_color,
                    edgecolor="black", linestyle=self.outline_style,
                    linewidth=self.outline_width)

        self.t = ax.text(self.position_x + self.box_indent, self.position_y, self.name,
                         va="center", fontweight="semibold", color="black", #font="monospace",
                         bbox=bbox)

    def get_text_start(self):
        return self.graph_box_start

    def get_text_end(self):
        return self.graph_box_end

    def calculate_bbox_dimensions(self, ax, graph_width):
        """
        Method that gets dimensions of plotted text box
        """
        transf = ax.transAxes.inverted()
        bb = self.t.get_window_extent(renderer=plt.gcf().canvas.get_renderer())
        bbox = bb.transformed(transf)

        TEXT_PADDING_IN_FIG_UNITS = 0.08
        box_padding = TEXT_PADDING_IN_FIG_UNITS / graph_width
        self.graph_box_start = bbox.x0 - box_padding
        self.graph_box_end = bbox.x1 + box_padding


import unittest

from mcstasscript.data.data import McStasMetaData


class TestMcStasMetaData(unittest.TestCase):
    """
    Various tests of McStasMetaData class
    """

    def test_McStasMetaData_return_type(self):
        """
        Test that newly created McStasMetaData has correct type
        """
        meta_data = McStasMetaData()
        self.assertIsInstance(meta_data, McStasMetaData)

    def test_McStasMetaData_init(self):
        """
        Test that newly created McStasMetaData has no content
        """
        meta_data = McStasMetaData()
        self.assertEqual(len(meta_data.info), 0)

    def test_McStasMetaData_add_info_len(self):
        """
        Test that info can be added to McStasMetaData
        """
        meta_data = McStasMetaData()
        meta_data.add_info("test", 3)
        self.assertEqual(len(meta_data.info), 1)

    def test_McStasMetaData_add_info(self):
        """
        Test that info can be read from McStasMetaData
        """
        meta_data = McStasMetaData()
        meta_data.add_info("test", 3)
        self.assertEqual(meta_data.info["test"], 3)

    def test_McStasMetaData_add_info_title(self):
        """
        Test that title can be set
        """
        meta_data = McStasMetaData()
        meta_data.set_title("title_test")
        self.assertEqual(meta_data.title, "title_test")

    def test_McStasMetaData_add_info_xlabel(self):
        """
        Test that xlabel can be set
        """
        meta_data = McStasMetaData()
        meta_data.set_xlabel("xlabel test")
        self.assertEqual(meta_data.xlabel, "xlabel test")

    def test_McStasMetaData_add_info_ylabel(self):
        """
        Test that ylabel can be set
        """
        meta_data = McStasMetaData()
        meta_data.set_ylabel("ylabel test")
        self.assertEqual(meta_data.ylabel, "ylabel test")

    def test_McStasMetaData_long_read_1d(self):
        """
        Test that extract_info can read appropriate info, 1d case
        """
        meta_data = McStasMetaData()
        meta_data.add_info("type", "array_1d(500)")
        meta_data.add_info("component", "test_A COMP")
        meta_data.add_info("filename", "test_A name")
        meta_data.add_info("xlimits", " 0.92 3.68")
        meta_data.add_info("xlabel", "test A xlabel")
        meta_data.add_info("ylabel", "test A ylabel")
        meta_data.add_info("title", "test A title")

        meta_data.extract_info()  # Converts info to attributes

        self.assertIsInstance(meta_data.dimension, int)
        self.assertEqual(meta_data.dimension, 500)
        self.assertIs(meta_data.component_name, "test_A COMP")
        self.assertIs(meta_data.filename, "test_A name")
        self.assertEqual(len(meta_data.limits), 2)
        self.assertEqual(meta_data.limits[0], 0.92)
        self.assertEqual(meta_data.limits[1], 3.68)
        self.assertIs(meta_data.xlabel, "test A xlabel")
        self.assertIs(meta_data.ylabel, "test A ylabel")
        self.assertIs(meta_data.title, "test A title")

    def test_McStasMetaData_long_read_2d(self):
        """
        Test that extract_info can read appropriate info, 2d case
        """
        meta_data = McStasMetaData()
        meta_data.add_info("type", "array_2d(500, 12)")
        meta_data.add_info("component", "test_A_COMP")
        meta_data.add_info("filename", "test_A_name")
        meta_data.add_info("xlimits", "-2.4 5.99 0.92 3.68")
        meta_data.add_info("xlabel", "test A xlabel")
        meta_data.add_info("ylabel", "test A ylabel")
        meta_data.add_info("title", "test A title")

        meta_data.extract_info()  # Converts info to attributes

        self.assertEqual(len(meta_data.dimension), 2)
        self.assertEqual(meta_data.dimension[0], 500)
        self.assertEqual(meta_data.dimension[1], 12)
        self.assertIs(meta_data.component_name, "test_A_COMP")
        self.assertIs(meta_data.filename, "test_A_name")
        self.assertEqual(len(meta_data.limits), 4)
        self.assertEqual(meta_data.limits[0], -2.4)
        self.assertEqual(meta_data.limits[1], 5.99)
        self.assertEqual(meta_data.limits[2], 0.92)
        self.assertEqual(meta_data.limits[3], 3.68)
        self.assertIs(meta_data.xlabel, "test A xlabel")
        self.assertIs(meta_data.ylabel, "test A ylabel")
        self.assertIs(meta_data.title, "test A title")


if __name__ == '__main__':
    unittest.main()



import unittest
import unittest.mock
import io
import numpy as np

import matplotlib
matplotlib.use('Agg')  # Set the backend to Agg

from mcstasscript.jb_interface.plot_interface import PlotInterface
from mcstasscript.jb_interface.plot_interface import LogCheckbox
from mcstasscript.jb_interface.plot_interface import ColormapDropdown
from mcstasscript.jb_interface.plot_interface import OrdersOfMagField

from mcstasscript.data.data import McStasMetaData
from mcstasscript.data.data import McStasDataBinned

import ipywidgets as widgets

def set_dummy_MetaDataBinned_1d():
    """
    Sets up simple McStasMetaData object with dimension, 1d case
    """
    meta_data = McStasMetaData()
    meta_data.component_name = "component for 1d"
    meta_data.dimension = 50

    meta_data.limits = [0, 1]
    meta_data.xlabel = ""
    meta_data.ylabel = ""
    meta_data.title = ""
    meta_data.filename = "dummy"

    return meta_data


def set_dummy_McStasDataBinned_1d():
    """
    Sets up simple McStasData object, 1d case
    """
    meta_data = set_dummy_MetaDataBinned_1d()

    intensity = np.arange(20)
    error = 0.5 * np.arange(20)
    ncount = 2 * np.arange(20)
    axis = np.arange(20)*5.0

    return McStasDataBinned(meta_data, intensity, error, ncount, xaxis=axis)


def set_dummy_MetaDataBinned_2d():
    """
    Sets up simple McStasMetaData object with dimensions, 2d case
    """
    meta_data = McStasMetaData()
    meta_data.component_name = "test a component"
    meta_data.dimension = [5, 4]

    meta_data.limits = [0, 1, 0, 1]
    meta_data.xlabel = ""
    meta_data.ylabel = ""
    meta_data.title = ""
    meta_data.filename = "dummy"

    return meta_data


def set_dummy_McStasDataBinned_2d():
    """
    Sets up simple McStasData object, 2d case
    """
    meta_data = set_dummy_MetaDataBinned_2d()

    intensity = np.arange(20).reshape(4, 5)
    error = 0.5 * np.arange(20).reshape(4, 5)
    ncount = 2 * np.arange(20).reshape(4, 5)

    return McStasDataBinned(meta_data, intensity, error, ncount)


def fake_data():
    return [set_dummy_McStasDataBinned_1d(),
            set_dummy_McStasDataBinned_2d(),
            set_dummy_McStasDataBinned_2d()]


class FakeChange:
    def __init__(self, new=None, old=None, name=None):
        self.new = new
        self.old = old
        self.name = name


class TestPlotInterface(unittest.TestCase):
    """
    Test of PlotInterface, mainly the set functions.
    Each set function does run the main update_plot function, ensuring
    this does not throw obvious errors under a mixed set of circumstances.
    The update_plot function is not tested directly.
    """
    def test_initialization_without_data(self):
        """
        Initialize PlotterInterface without arguments
        """

        interface = PlotInterface()
        self.assertEqual(interface.data, None)

    def test_initialization_with_data(self):
        """
        Initialize PlotterInterface with data
        """
        data = fake_data()
        interface = PlotInterface(data)

        self.assertEqual(len(interface.data), 3)
        self.assertEqual(interface.data[0].Intensity[5], 5)
        self.assertEqual(interface.data[1].Intensity[2, 3], 13)

    def test_show_interface_return(self):
        """
        Ensure the show_interface method returns a widget of type HBox
        """
        interface = PlotInterface()
        widget = interface.show_interface()

        self.assertIsInstance(widget, widgets.widgets.widget_box.HBox)

    def test_set_data(self):
        """
        Initialize PlotterInterface without data, add later and ensure
        monitor_dropdown received the data and applied it to widget options.

        Furthermore the widget options should be made unique as there are
        two identical monitors in the fake data.
        """

        interface = PlotInterface()
        self.assertEqual(interface.data, None)

        interface.show_interface() # Needed to set up monitor_dropdown

        interface.set_data(fake_data())
        self.assertEqual(len(interface.data), 3)
        self.assertEqual(interface.data[0].Intensity[5], 5)
        self.assertEqual(interface.data[1].Intensity[2, 3], 13)

        mon_drop = interface.monitor_dropdown
        self.assertEqual(len(mon_drop.data), 3)
        self.assertEqual(mon_drop.data[0].Intensity[5], 5)
        self.assertEqual(mon_drop.data[1].Intensity[2, 3], 13)

        self.assertEqual(mon_drop.widget.options[0], "component for 1d")
        self.assertEqual(mon_drop.widget.options[1], "test a component")

        # The last monitor is identical to the second, so the name is modified
        self.assertEqual(mon_drop.widget.options[2], "test a component_1")

    def test_set_current_monitor(self):
        """
        Check monitor can be set and that error occurs if wrong name given
        """

        interface = PlotInterface(fake_data())
        interface.show_interface()

        interface.set_current_monitor("test a component")
        self.assertEqual(interface.current_monitor, "test a component")

        with self.assertRaises(NameError):
            interface.set_current_monitor("component that doesnt exist")

    def test_set_log_mode(self):
        """
        Check that set_log_mode works, even through widget
        """

        interface = PlotInterface()
        interface.show_interface()

        self.assertFalse(interface.log_mode)
        interface.set_log_mode(True)
        self.assertTrue(interface.log_mode)

        log_checkbox = LogCheckbox(interface.log_mode, interface.set_log_mode)

        fake_change = FakeChange(new=False)

        log_checkbox.update(fake_change)
        self.assertFalse(interface.log_mode)

    def test_set_orders_of_mag(self):
        """
        Check that set orders of mag works, even through widget
        """
        interface = PlotInterface()
        interface.show_interface()

        interface.set_orders_of_mag(37)
        self.assertEqual(interface.orders_of_mag, 37)

        log_orders_of_mag = OrdersOfMagField(interface.set_orders_of_mag)

        fake_change = FakeChange(new=42)

        log_orders_of_mag.update(fake_change)
        self.assertEqual(interface.orders_of_mag, 42)

    def test_set_colormap(self):
        """
        Check that set_colormap works, even through widget
        """
        interface = PlotInterface()
        interface.show_interface()

        interface.set_colormap("hot")
        self.assertEqual(interface.colormap, "hot")

        colormap_dropdown = ColormapDropdown(interface.set_colormap)

        fake_change = FakeChange(new="Purples")
        colormap_dropdown.update_cmap(fake_change)
        self.assertEqual(interface.colormap, "Purples")








import os
import math
import re


def remove_c_comments(code):
    """
    Removes comments from a multiline piece of c code
    """
    # Remove single-line comments
    code = re.sub(r'//.*', '', code)

    # Remove multi-line comments
    code = re.sub(r'/\*.*?\*/', '', code, flags=re.DOTALL)

    # Remove empty lines
    code = '\n'.join([line for line in code.split('\n') if line.strip()])

    return code


def c_integer_literal_base(s: str) -> int:
    """
    Determine the base of a C style integer literal.

    base    literal
    ------- --------------------
    binary      0b### for # in (0,1)
    octal       0#### for # in (0,7)
    decimal     N#### for N in (1,9) and # in (0,9)
    hexadecimal 0x### for # in (0,9)&(A,F)

    The ambiguity between octal and decimal for '0' is not important
    since both result in the same integer value.
    """
    if len(s) == 0:
        return 0
    if '0' != s[0]:
        return 10
    if 'x' in s:
        return 16
    if 'b' in s:
        return 2
    return 8


class ComponentInfo:
    """
    Internal class used to store information on parameters of components
    """

    def __init__(self):
        self.name = ""
        self.category = ""
        self.parameter_names = []
        self.parameter_defaults = {}
        self.parameter_types = {}
        self.parameter_comments = {}
        self.parameter_units = {}


class ComponentReader:
    """
    Class for retrieving information on available McStas components

    Recursively reads all component files in hardcoded list of
    folders that represents the component categories in McStas.
    The results are stored in a dictionary with ComponentInfo
    instances, the keys are the names of the components. After
    the components in the McStas installation are read, any
    components present in the current work directory is read,
    and these will overwrite existing information, consistent
    with how McStas reads component definitions.

    """

    def __init__(self, mcstas_path, input_path="."):
        """
        Reads all component files in standard folders. Recursive, so
        subfolders of these folders are included.

        Parameters
        ----------
        mcstas_path : str
            Path to McStas folder, used to find the installed components

        keyword arguments:
            input_path : str
                Path to work directory, most often current directory

        """

        # Hardcoded whitelist of foldernames
        folder_list = ["sources",
                       "optics",
                       "samples",
                       "monitors",
                       "misc",
                       "contrib",
                       "obsolete",
                       "union",
                       "astrox"]

        self.component_path = {}
        self.component_category = {}

        for folder in folder_list:
            abs_path = os.path.join(mcstas_path, folder)
            abs_path = os.path.abspath(abs_path)
            self._find_components(abs_path)

        # Will overwrite McStas components with definitions in input_folder
        current_directory = os.getcwd()

        # Set up absolute input_path
        if os.path.isabs(input_path):
            input_directory = input_path
        else:
            if input_path == ".":
                # Default case, avoid having /./ in absolute path
                input_directory = current_directory
            else:
                input_directory = os.path.join(current_directory, input_path)

        if not os.path.isdir(input_directory):
            print("input_path: ", input_directory)
            raise ValueError("Can't find given input_path,"
                             + " directory must exist.")

        """
        If components are present both in the McStas install and the
        work directory, the version in the work directory is used. The user
        is informed of this behavior when the instrument object is created.
        """

        self.load_components_from_folder(input_directory, "work directory")

    def load_components_from_folder(self, folder, name, verbose=True):
        """
        Loads McStas components from given absolute path

        folder : (str) Path for folder to search for components in
        name : (str) Used for displaying help messages about these components
        verbose : (bool) If True, help messages are shown about the process
        """

        if not os.path.isdir(folder):
            if verbose:
                print("Did not find specified folder: " + folder)
            return

        overwritten_components = []
        for file in os.listdir(folder):
            if file.endswith(".comp"):
                abs_path = os.path.join(folder, file)
                component_name = os.path.split(abs_path)[1].split(".")[-2]

                if component_name in self.component_path:
                    overwritten_components.append(file)

                self.component_path[component_name] = abs_path
                self.component_category[component_name] = name

        # Report components found in the work directory and install to the user
        if len(overwritten_components) > 0 and verbose:
            print(f"The following components are found in the {name}"
                  + " / input_path:")
            for component_name in overwritten_components:
                print("    ", component_name)

            print("These definitions will be used instead of the installed "
                  + "versions.")

    def show_categories(self):
        """
        Method that will show all component categories available

        Sorted alphabetically for easier readability and consistency
        """
        categories = []
        for component, category in self.component_category.items():
            if category not in categories:
                categories.append(category)

        categories.sort()
        for category in categories:
            print(" " + category)

    def show_components_in_category(self, category_input, **kwargs):
        """
        Method that will show all components in given category

        """
        if "line_length" in kwargs:
            line_limit = int(kwargs["line_length"])
            if line_limit < 20:
                raise ValueError("line_length should be more than 20 "
                                 + "characters, was " + str(line_limit))
        else:
            line_limit = 100

        empty_category = True
        to_print = []
        for component, category in self.component_category.items():
            if category == category_input:
                to_print.append(component)
                empty_category = False

        to_print.sort()
        if empty_category:
            print("No components found in this category! "
                  + "Available categories:")
            self.show_categories()

        elif len(to_print) < 10:
            for component in to_print:
                print(" " + component)
        else:
            # Prints in columns, maximum 4 and maximum line length line_limit
            columns = 5
            total_line_length = 1000
            while(total_line_length > line_limit):
                columns = columns - 1

                c_length = math.ceil(len(to_print)/columns)
                last_length = len(to_print) - (columns-1)*c_length

                column = []
                longest_name = []
                for col in range(0, columns-1):
                    current_list = to_print[c_length*col:c_length*(col+1)]
                    column.append(current_list)
                    longest_name.append(len(max(current_list, key=len)))

                column.append(to_print[c_length*(columns-1):])
                longest_name.append(len(max(column[columns-1], key=len)))

                total_line_length = 1 + sum(longest_name) + (columns-1)*3

            for line_nr in range(c_length):
                print(" ", end="")
                for col in range(columns-1):
                    this_name = column[col][line_nr]
                    print(this_name
                          + " "*(longest_name[col] - len(this_name))
                          + "   ", end="")  # More columns left, dont break
                if line_nr < last_length:
                    this_name = column[columns-1][line_nr]
                    print(this_name)
                else:
                    print("")

    def load_all_components(self):
        """
        Method that loads information on all components into memory.

        """

        return_dict = {}
        for comp_name, abs_path in self.component_path.items():
            return_dict[comp_name] = self.read_component_file(abs_path)

        return return_dict

    def read_name(self, component_name):
        """
        Returns ComponentInfo of component with name component_name.

        Uses table of absolute paths to all known components, and
        reads the appropriate file in order to generate the information.

        """

        if component_name not in self.component_path:
            raise NameError("No component named "
                            + component_name
                            + " in McStas installation or "
                            + "current work directory.")

        output = self.read_component_file(self.component_path[component_name])

        # Category loaded using path, in case of Work directory it fails
        if self.component_category[component_name] == "work directory":
            output.category = "work directory"  # Corrects category

        return output

    def _find_components(self, absolute_path):
        """
        Recursive read function, can read either file or entire folder

        Updates the component_info_dict with the findings that are
        stored as ComoponentInfo instances.

        """

        if not os.path.isabs(absolute_path):
            raise RuntimeError("_find_components received non absolute path")

        if not os.path.isdir(absolute_path):
            if absolute_path.endswith(".comp"):
                # read this file
                component_name = os.path.split(absolute_path)[1].split(".")[-2]
                self.component_path[component_name] = absolute_path

                head = os.path.split(absolute_path)[0]
                component_category = os.path.split(head)[1]
                self.component_category[component_name] = component_category
        else:
            for file in os.listdir(absolute_path):
                absolute_file_path = os.path.join(absolute_path, file)
                self._find_components(absolute_file_path)

    def read_component_file(self, absolute_path):
        """
        Reads a component file and expands component_info_dict

        The information is stored as ComponentClass instances.
        """

        result = ComponentInfo()

        file_o = open(absolute_path, "r")

        line_number = 0
        while True:
            line_number += 1
            line = file_o.readline()

            if not line:
                # Exit at end of file
                break

            # find parameter comments
            if line.startswith("* %P"):

                while True:
                    this_line = file_o.readline()

                    if this_line.startswith("DEFINE COMPONENT"):
                        # No more comments to read through
                        break

                    if ":" in this_line:
                        tokens = this_line.split(":")

                        variable_name = tokens[0]
                        variable_name = variable_name.replace("*", "")
                        variable_name = variable_name.strip()
                        if " " in variable_name:
                            name_tokens = variable_name.split(" ")
                            variable_name = name_tokens[0]

                        if len(tokens[1]) > 2:
                            comment = tokens[1].strip()

                            if "[" in comment:  # Search for unit
                                # If found, store it and remove from string
                                unit = comment[comment.find("[") + 1:
                                               comment.find("]")]
                                result.parameter_units[variable_name] = unit
                                comment = comment[comment.find("]") + 1:]
                                comment = comment.strip()

                            # Store the comment
                            result.parameter_comments[variable_name] = comment
                    elif "[" in this_line and "]" in this_line:
                        tokens = this_line.split("[")

                        variable_name = tokens[0]
                        variable_name = variable_name.replace("*", "")
                        variable_name = variable_name.strip()

                        unit = this_line[this_line.find("[") + 1:
                                         this_line.find("]")]
                        result.parameter_units[variable_name] = unit

                        comment = this_line[this_line.find("]") + 1:]
                        comment = comment.strip()
                        result.parameter_comments[variable_name] = comment

            # find definition parameters and their values
            if (line.strip().startswith("DEFINITION PARAMETERS")
                    or line.strip().startswith("SETTING PARAMETERS")):

                define_section = line
                while True:
                    line = file_o.readline()

                    end_keywords = ("SHARE", "INITIALIZE", "INITIALISE", "DECLARE", "TRACE", "DEPENDENCY")
                    if line.strip().upper().startswith(end_keywords) or not line:
                        break

                    define_section += line

                clean_define_section = remove_c_comments(define_section)

                # Define the delimiters as a list of strings
                delimiters = ["DEFINITION PARAMETERS", "SETTING PARAMETERS", "OUTPUT PARAMETERS"]

                # Create a regex pattern using alternation and join the delimiters with the '|' symbol
                delimiter_pattern = r'\s*(' + '|'.join(map(re.escape, delimiters)) + r')\s*'

                # Split the text using pattern
                clean_define_sections = re.split(delimiter_pattern, clean_define_section)

                # Extract parameters from definition and settings part
                parameter_section = ""
                for index, section in enumerate(clean_define_sections):
                    if section in ("DEFINITION PARAMETERS", "SETTING PARAMETERS"):
                        parameter_section += clean_define_sections[index + 1].strip("(").strip(")") + ", "

                # Convert parameter section to single line, then split in parts separated by comma
                parameter_section = parameter_section.replace('\n', ' ')
                parameter_parts = parameter_section.split(",")
                # Combine parts that should be together, for example by brackets
                parameter_parts = self.correct_for_brackets(parameter_parts)

                # Each part now corresponds to a parameter to be read
                for part in parameter_parts:

                    temp_par_type = "double"

                    part = part.strip()

                    possible_declare = part.split(" ")
                    possible_type = possible_declare[0].strip()
                    if "int" == possible_type:
                        temp_par_type = "int"
                        # remove int from part
                        part = "".join(possible_declare[1:])
                    if "string" == possible_type:
                        temp_par_type = "string"
                        # remove string from part
                        part = "".join(possible_declare[1:])
                    if "double" == possible_type:
                        temp_par_type = "double"
                        # remove double from part
                        part = "".join(possible_declare[1:])
                    if "vector" == possible_type:
                        temp_par_type = "double"
                        # remove double from part
                        part = "".join(possible_declare[1:])

                    part = part.replace(" ", "")
                    if part == "":
                        continue

                    if "=" not in part:
                        # no default value, required parameter
                        result.parameter_names.append(part)
                        result.parameter_defaults[part] = None
                        result.parameter_types[part] = temp_par_type
                    else:
                        # default value available
                        name_value = part.split("=")
                        par_name = name_value[0].strip()
                        par_value = name_value[1].strip()

                        if temp_par_type == "double":
                            try:
                                par_value = float(par_value)
                            except ValueError:
                                # value could be parameter name
                                par_value = par_value
                        elif temp_par_type == "int":
                            par_value = int(par_value, c_integer_literal_base(par_value))

                        result.parameter_names.append(par_name)
                        result.parameter_defaults[par_name] = par_value
                        result.parameter_types[par_name] = temp_par_type

                # End while loop running through file when parameters are read
                break

        file_o.close()

        result.name = os.path.split(absolute_path)[1].split(".")[-2]

        tail = os.path.split(absolute_path)[0]
        result.category = os.path.split(tail)[1]

        """
        To lower memory use one could remove all comments and units that
        does not correspond to a found parameter name.
        """

        return result

    def correct_for_brackets(self, parameter_parts):
        """
        Given list of string elements, correct for brackets will
        combine terms until curly brackets are balanced, for example:

        ["A", "{B", "C", "D}", "E"] would return ["A", "{B,C,D}", "E"]

        Default values of vectors can be given in such a manner in
        McStas components, and without this each part would be recognized
        as different parameters.
        """
        corrected_parts = []
        index = 0
        while True:

            current_part = parameter_parts[index]
            inner_index = 0
            while True:
                if current_part.count("{") == current_part.count("}"):
                    corrected_parts.append(current_part)
                    index += inner_index
                    break
                else:
                    inner_index += 1
                    current_part += "," + parameter_parts[index+inner_index]

            index += 1

            if index >= len(parameter_parts):
                break

        return corrected_parts



def has_component(instrument, component_name=None, component_type=None):
    """
    Checks if given instrument has a component with given name and type

    instrument : McStasScript instrument object

    component_name : (str) name of component

    component_type : (str) type of component, for example Arm
    """
    if component_name is None and component_type is None:
        raise ValueError("Specify component_name, component_type or both.")

    # Run through components in instrument
    for comp in instrument.component_list:
        if component_name is not None:
            if comp.name != component_name:
                continue

        if component_type is not None:
            if comp.component_name != component_type:
                continue

        # Found case where both component name and type matches
        return True

    # Did not find any such component
    return False


def has_parameter(instrument, parameter_name, parameter_type=None):
    """
    Checks if instrument have parameter with given name and type

    instrument : McStasScript instrument object

    parameter_name : (str) Name of parameter

    parameter_type : (str) Type of parameter, "" treated as double
    """
    if parameter_name is None and parameter_type is None:
        raise ValueError("Specify parameter_name, parameter_type or both.")

    parameter_dict = instrument.parameters.parameters

    if parameter_name not in parameter_dict:
        return False

    if parameter_type is not None:
        found_type = parameter_dict[parameter_name].type

        if parameter_type == "double" and found_type == "":
            # Default McStas type is double
            return True

        if found_type != parameter_type:
            return False

    return True


def all_parameters_set(instrument):
    """
    Checks if all parameters of given instrument have default values

    instrument : McStasScript instrument object
    """
    parameter_dict = instrument.parameters.parameters

    for par_object in parameter_dict.values():
        if par_object.value is None:
            return False

    return True


import unittest

from mcstasscript.data.data import McStasPlotOptions


class TestMcStasPlotOptions(unittest.TestCase):
    """
    Various test of McStasPlotOptions class
    """

    def test_McStasPlotOptions_default_log(self):
        """
        Test that newly created McStasPlotOptions log attribute
        has correct type and default value
        """
        plot_options = McStasPlotOptions()
        self.assertIsInstance(plot_options.log, bool)
        self.assertFalse(plot_options.log)

    def test_McStasPlotOptions_default_orders_of_mag(self):
        """
        Test that newly created McStasPlotOptions orders_of_mag
        has the correct default value
        """
        plot_options = McStasPlotOptions()
        self.assertEqual(plot_options.orders_of_mag, 300)

    def test_McStasPlotOptions_default_colormap(self):
        """
        Test that newly created McStasPlotOptions colormap has
        the correct default value
        """
        plot_options = McStasPlotOptions()
        self.assertIs(plot_options.colormap, "jet")

    def test_McStasPlotOptions_default_show_colorbar(self):
        """
        Test that newly created McStasPlotOptions has correct
        default value for show_colorbar
        """
        plot_options = McStasPlotOptions()
        self.assertIs(plot_options.show_colorbar, True)

    def test_McStasPlotOptions_default_cut_max(self):
        """
        Test that newly created McStasPlotOptions has correct
        default value for cut_max
        """
        plot_options = McStasPlotOptions()
        self.assertIs(plot_options.cut_max, 1)

    def test_McStasPlotOptions_default_cut_min(self):
        """
        Test that newly created McStasPlotOptions has correct
        default value for cut_min
        """
        plot_options = McStasPlotOptions()
        self.assertIs(plot_options.cut_min, 0)

    def test_McStasPlotOptions_default_x_axis_multiplier(self):
        """
        Test that newly created McStasPlotOptions has correct
        default value for x_axis_multiplier
        """
        plot_options = McStasPlotOptions()
        self.assertIs(plot_options.x_limit_multiplier, 1)

    def test_McStasPlotOptions_default_y_axis_multiplier(self):
        """
        Test that newly created McStasPlotOptions has correct
        default value for y_axis_multiplier
        """
        plot_options = McStasPlotOptions()
        self.assertIs(plot_options.y_limit_multiplier, 1)

    def test_McStasPlotOptions_default_top_lim(self):
        """
        Test that newly created McStasPlotOptions has correct
        default value for top_lim
        """
        plot_options = McStasPlotOptions()
        self.assertIs(plot_options.custom_ylim_top, False)

    def test_McStasPlotOptions_default_bottom_lim(self):
        """
        Test that newly created McStasPlotOptions has correct
        default value for left_lim
        """
        plot_options = McStasPlotOptions()
        self.assertIs(plot_options.custom_ylim_bottom, False)

    def test_McStasPlotOptions_default_left_lim(self):
        """
        Test that newly created McStasPlotOptions has correct
        default value for left_lim
        """
        plot_options = McStasPlotOptions()
        self.assertIs(plot_options.custom_xlim_left, False)

    def test_McStasPlotOptions_default_right_lim(self):
        """
        Test that newly created McStasPlotOptions has correct
        default value for right_lim
        """
        plot_options = McStasPlotOptions()
        self.assertIs(plot_options.custom_xlim_right, False)

    def test_McStasPlotOptions_set_log(self):
        """
        Test that set_options works on log parameter which
        can be set both with an integer and a bool.
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(log=True)
        self.assertIsInstance(plot_options.log, bool)
        self.assertTrue(plot_options.log)

        plot_options.set_options(log=0)
        self.assertIsInstance(plot_options.log, bool)
        self.assertFalse(plot_options.log)

        plot_options.set_options(log=1)
        self.assertIsInstance(plot_options.log, bool)
        self.assertTrue(plot_options.log)

    def test_McStasPlotOptions_set_orders_of_mag(self):
        """
        Check that set_options works with orders_of_mag keyword
        argument
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(orders_of_mag=5.2)
        self.assertEqual(plot_options.orders_of_mag, 5.2)

    def test_McStasPlotOptions_set_colormap(self):
        """
        Check that set_options work with colormap keyword argument
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(colormap="hot")
        self.assertIs(plot_options.colormap, "hot")

    def test_McStasPlotOptions_set_show_colorbar(self):
        """
        Check that set_options work with show_colormap keyword
        argument
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(show_colorbar=False)
        self.assertIs(plot_options.show_colorbar, False)

    def test_McStasPlotOptions_set_cut_max(self):
        """
        Check that set_options work with cut_max keyword
        argument
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(cut_max=0.8)
        self.assertIs(plot_options.cut_max, 0.8)

    def test_McStasPlotOptions_set_cut_min(self):
        """
        Check that set_options work with cut_min keyword
        argument
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(cut_min=0.2)
        self.assertIs(plot_options.cut_min, 0.2)

    def test_McStasPlotOptions_set_x_axis_multiplier(self):
        """
        Check that set_options work with x_axis_multiplier
        keyword argument
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(x_axis_multiplier=2.8)
        self.assertIs(plot_options.x_limit_multiplier, 2.8)

    def test_McStasPlotOptions_set_y_axis_multiplier(self):
        """
        Check that set_options work with y_axis_multiplier
        keyword argument
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(y_axis_multiplier=0.1)
        self.assertIs(plot_options.y_limit_multiplier, 0.1)

    def test_McStasPlotOptions_set_top_lim(self):
        """
        Check that set_options work with top_lim keyword argument
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(top_lim=128.9)
        self.assertIs(plot_options.custom_ylim_top, True)
        self.assertIs(plot_options.top_lim, 128.9)

    def test_McStasPlotOptions_set_bottom_lim(self):
        """
        Check that set_options work with bottom_lim keyword
        argument
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(bottom_lim=120.9)
        self.assertIs(plot_options.custom_ylim_bottom, True)
        self.assertIs(plot_options.bottom_lim, 120.9)

    def test_McStasPlotOptions_set_left_lim(self):
        """
        Check that set_options work with left_lim keyword argument
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(left_lim=9.2)
        self.assertIs(plot_options.custom_xlim_left, True)
        self.assertIs(plot_options.left_lim, 9.2)

    def test_McStasPlotOptions_set_right_lim(self):
        """
        Check that set_options work with right_lim keyword argument
        """
        plot_options = McStasPlotOptions()
        plot_options.set_options(right_lim=1.4)
        self.assertIs(plot_options.custom_xlim_right, True)
        self.assertIs(plot_options.right_lim, 1.4)


if __name__ == '__main__':
    unittest.main()



import math

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

from mcstasscript.helper.plot_helper import _fmt
from mcstasscript.helper.plot_helper import _find_min_max_I
from mcstasscript.helper.plot_helper import _plot_fig_ax
from mcstasscript.helper.plot_helper import _handle_kwargs


def make_plot(data_list, **kwargs):
    """
    make_plot plots contents of McStasData objects given in list

    Here a new figure is used for each dataset

    Plotting is controlled through options assosciated with the
    McStasData objects.  If a list is given, the plots appear in one
    subplot.
    """

    data_list = _handle_kwargs(data_list, **kwargs)

    if "figsize" in kwargs:
        figsize = kwargs["figsize"]
        if isinstance(figsize, list):
            figsize = (figsize[0], figsize[1])
    else:
        figsize = (13, 7)

    for data in data_list:
        fig, ax0 = plt.subplots(figsize=figsize, tight_layout=True)
        _plot_fig_ax(data, fig, ax0, **kwargs)

    if "filename" in kwargs:
        fig.tight_layout()
        fig.savefig(kwargs["filename"])
    else:
        plt.show()


def make_sub_plot(data_list, **kwargs):
    """
    make_sub_plot plots contents of McStasData objects given in list

    It is fit into one big figure, each data set as a subplot.

    Plotting is controlled through options assosciated with the
    McStasData objects.  If a list is given, the plots appear in one
    subplot.
    """

    data_list = _handle_kwargs(data_list, **kwargs)

    number_of_plots = len(data_list)
    if number_of_plots == 0:
        print("No data to plot")
        return

    # Find reasonable grid size for the number of plots
    special_cases = {
        1: (1, 1),
        4: (2, 2),
    }

    if number_of_plots in special_cases:
        dim1 = special_cases[number_of_plots][0]
        dim2 = special_cases[number_of_plots][0]
    else:
        if number_of_plots < 3:
            dim2 = number_of_plots
            dim1 = 1
        else:
            dim2 = 3
            dim1 = math.ceil(number_of_plots / dim2)

    if "figsize" in kwargs:
        figsize = kwargs["figsize"]
        if isinstance(figsize, list):
            figsize = (figsize[0], figsize[1])
    else:
        # Adjust figure size after number of plots
        figsize = (1 + dim2*4, 0.5 + 3.0*dim1)
        if dim1 == 1 and dim2 == 1:
            # Single plots can be a bit larger
            figsize = (7, 5)

    fig, axs = plt.subplots(dim1, dim2, figsize=figsize, tight_layout=True)
    axs = np.array(axs)
    ax = axs.reshape(-1)

    for data, ax0 in zip(data_list, ax):
        _plot_fig_ax(data, fig, ax0, **kwargs)

    fig.tight_layout()

    if "filename" in kwargs:
        fig.tight_layout()
        fig.savefig(kwargs["filename"])
        plt.close(fig)
    else:
        plt.show()


def make_animation(data_list, **kwargs):
    """
    Creates an animation from list of McStasData objects

    Parameters
    ----------
    data_list : list of McStasData
        List of McStasData objects for animation

    Keyword arguments
    -----------------
        filename : str
            Filename for saving the gif

        fps : float
            Number of frames per second

    """

    data_list = _handle_kwargs(data_list, **kwargs)

    if "figsize" in kwargs:
        figsize = kwargs["figsize"]
        if isinstance(figsize, list):
            figsize = (figsize[0], figsize[1])
    else:
        figsize = (13, 7)

    if "fps" in kwargs:
        period_in_ms = 1000 / kwargs["fps"]
    else:
        period_in_ms = 200

    # find limits for entire dataset
    maximum_values = []
    minimum_values = []

    is_1D = False
    is_2D = False

    for data in data_list:
        if isinstance(data.metadata.dimension, int):
            is_1D = True

        elif len(data.metadata.dimension) == 2:
            is_2D = True

        min_value, max_value = _find_min_max_I(data)

        # When data empty, min and max value is 0, skip
        if not (min_value == 0 and max_value == 0):
            minimum_values.append(min_value)
            maximum_values.append(max_value)

    if is_1D and is_2D:
        raise ValueError(
            "Both 1D and 2D data in animation, only one allowed.")

    if len(minimum_values) == 0:
        raise ValueError(
            "No data found for animation!")

    maximum_value = np.array(maximum_values).max()
    minimum_value = np.array(minimum_values).min()

    if "orders_of_mag" in kwargs:
        orders_of_mag = kwargs["orders_of_mag"]
        mag_diff = np.log10(maximum_value) - np.log10(minimum_value)
        if mag_diff > orders_of_mag:
            minimum_value_log10 = np.log10(maximum_value) - orders_of_mag
            minimum_value = 10**(minimum_value_log10)

    kwargs["fixed_minimum_value"] = minimum_value
    kwargs["fixed_maximum_value"] = maximum_value

    fig, ax0 = plt.subplots(figsize=figsize, tight_layout=True)
    im = _plot_fig_ax(data_list[0], fig, ax0, **kwargs)

    def animate_2D(index):
        data = data_list[index]
        intensity = data.Intensity

        im.set_array(intensity.ravel())
        return im,

    anim = animation.FuncAnimation(fig, animate_2D,
                                   frames=len(data_list),
                                   interval=period_in_ms,
                                   blit=False, repeat=True)

    plt.show()

    # The animation doesn't play unless it is saved. Bug.
    if "filename" in kwargs:
        filename = kwargs["filename"]
        if not filename.endswith(".gif"):
            filename = filename + ".gif"

        # check if imagemagick available?
        print("Saving animation with filename : \"" + filename + "\"")
        anim.save(filename, writer="imagemagick")



class SectionReader:
    """
    Super class for the many necessary readers
    """

    def __init__(self, Instr, write_file, product_filename,
                 get_next_line, return_line):
        self.Instr = Instr
        self.write_file = write_file
        self.product_filename = product_filename
        self.instr_name = ""
        self.get_next_line = get_next_line
        self.return_line = return_line

    def set_instr_name(self, name):
        self.instr_name = name

    def _write_to_file(self, string_array):
        """
        In case a py file is being written, this function writes to the
        appropriate file.
        """

        if self.write_file:
            with open(self.product_filename, "a") as product_file:
                for string in string_array:
                    product_file.write(string)

    def _kw_to_string(self, kwargs):
        """
        Used when a dict containing keyword arguments need to be written
        to a string. This string can be used as argument in method call.
        """

        output_string = ""
        for kwarg in kwargs:
            output_string += ", "
            output_string += kwarg + "=" + str(kwargs[kwarg])

        return output_string

    def _split_func(self, *args):
        """
        Returns list of strings seperated by commas that are not
        within open parenthesis.
        """

        string = args[0]
        split_character = args[1]

        if len(args) == 3:
            limit = args[2]
        else:
            limit = -1

        split_positions = []
        parenthesis = 0
        for index in range(len(string)):
            character = string[index]
            if (character == split_character and parenthesis == 0
                    and limit != 0):
                split_positions.append(index)
                limit -= 1
            else:
                if character == "(":
                    parenthesis += 1
                if character == ")":
                    parenthesis -= 1

        split_positions.append(len(string)+1)  # virtual comma at the end

        result = []
        last_position = 0
        for position in split_positions:
            result.append(string[last_position:position])
            last_position = position + 1

        return result

    def _split_func_brack(self, *args):
        """
        Returns list of strings seperated by commas that are not
        within open parenthesis / brackets
        """

        string = args[0]
        split_character = args[1]

        if len(args) == 3:
            limit = args[2]
        else:
            limit = -1

        split_positions = []
        parenthesis = 0
        brackets = 0
        for index in range(len(string)):
            character = string[index]
            if (character == split_character and parenthesis == 0
                    and brackets == 0 and limit != 0):
                split_positions.append(index)
                limit -= 1
            else:
                if character == "(":
                    parenthesis += 1
                if character == ")":
                    parenthesis -= 1
                if character == "{":
                    brackets += 1
                if character == "}":
                    brackets -= 1

        split_positions.append(len(string)+1)  # virtual comma at the end

        result = []
        last_position = 0
        for position in split_positions:
            result.append(string[last_position:position])
            last_position = position + 1

        return result

    def _in_func(self, string, character):
        """
        Returns true of character is in string when excluding occurances
        within parenthesis.
        """

        if len(self._split_func(string, character, 1)) == 2:
            return True
        else:
            return False

    def _in_func_brack(self, string, character):
        """
        Returns true of character is in string when excluding occurances
        within parenthesis and brackets.
        """

        if len(self._split_func_brack(string, character, 1)) == 2:
            return True
        else:
            return False



import copy

class DiagnosticsInstrument:
    def __init__(self, instr):
        """
        Diagnostics instrument object with related capabilities

        The diagnostics instrument takes a copy of an instrument object
        and can modify this version in ways relevant for diagnostics of
        the instrument. It can remove use of PREVIOUS and correct the
        use of target_index to make it safe to add additional components
        without altering the instrument. It carries it's own settings
        and parameters that wont impact the original instrument. The
        diagnostics instrument can be reset to the original state, and
        will have settings and parameters applied again.

        Parameters:

        instr : McCode_instr object
            Instrument which needs a diagnostics copy
        """
        self.original_instr = instr
        self.instr = None
        self.instr_settings = {}
        self.instr_parameters = {}
        self.reset_instr()

        self.component_list = self.instr.make_component_subset()

    def reset_instr(self):
        """
        Resets instrument to original state, with new pars and settings
        """
        self.instr = copy.deepcopy(self.original_instr)
        self.instr.settings(**self.instr_settings)
        self.instr.set_parameters(**self.instr_parameters)

    def settings(self, **kwargs):
        """
        Apply settings as in an instrument object
        """
        self.instr.settings(**kwargs)
        self.instr_settings.update(kwargs)

    def show_settings(self):
        """
        Show settings of diagnostics instrument
        """
        self.instr.show_settings()

    def set_parameters(self, **kwargs):
        """
        Set parameters as in instrument object
        """
        self.instr.set_parameters(**kwargs)
        self.instr_parameters.update(kwargs)

    def show_parameters(self):
        """
        Show parameters as on instrument object
        """
        self.instr.show_parameters()

    def remove_previous_use(self):
        """
        Replaces use of PREVIOUS with direct references
        """
        self.component_list = self.instr.make_component_subset()
        previous_component = None
        for comp in self.component_list:
            if comp.AT_reference == "PREVIOUS":
                comp.set_AT_RELATIVE(previous_component)

            if comp.ROTATED_specified:
                if comp.ROTATED_reference == "PREVIOUS":
                    comp.set_ROTATED_RELATIVE(previous_component)

            previous_component = comp

    def correct_target_index(self):
        """
        Corrects target_index based on original instrument
        """
        original_component_list = self.original_instr.make_component_subset()
        original_comp_names = [x.name for x in original_component_list]

        modified_component_list = self.instr.make_component_subset()
        modified_comp_names = [x.name for x in modified_component_list]

        for comp in original_component_list:
            # Find components that use target index

            if not hasattr(comp, "target_index"):
                # Component doesnt have the target_index setting
                continue
            if comp.target_index is None:
                # A value has not been specified for target_index setting
                continue
            if comp.target_index == 0:
                # target_index is disabled
                continue

            # Only here if target_index is used, correct it in modified instr

            # Find original index and the name of the original target
            original_comp_index = original_comp_names.index(comp.name)
            comp_target_name = original_comp_names[original_comp_index + int(comp.target_index)]

            # Find index of the original and target in modified instrument
            modified_comp_index = modified_comp_names.index(comp.name)
            index_of_new_target = modified_comp_names.index(comp_target_name)
            new_target_index = index_of_new_target - modified_comp_index

            # Apply the new target_index to the component in the modified instrument
            modified_comp = self.instr.get_component(comp.name)
            modified_comp.target_index = new_target_index



import numpy as np
import matplotlib.pyplot as plt

class PlotOverview:
    """
    Class for plotting list of views at different points in instrument
    """
    def __init__(self, event_plotter_list, view_list):
        """
        Stores list of event plotters and views, able to create figure

        parameters:

        event_plotter_list : list of EventPlotter objects
            EventPlotter objects for each investigated point in instrument

        view_list : list of View objects
            View specifications which will all be used for each point
        """

        self.event_plotter_list = event_plotter_list
        self.n_points = len(event_plotter_list)
        self.views = view_list
        self.n_plots = len(view_list)

    def plot_all(self, figsize=None, same_scale=True):
        """
        Plots all views for all guide points

        The same_scale features is enabled per default, but even then
        individual Views can opt out of being drawn on the same scale.

        Parameters:

        figsize : tuple of length 2
            size of figure, default scales according to number of plots

        same_scale : bool
            Allow the use of the same scale feature (default)
        """
        if same_scale:
            self.set_same_scale()

        if figsize is None:
            # Scale size after number of plots
            figsize = (1 + self.n_plots*3, self.n_points*3)

        fig, axs = plt.subplots(self.n_points, self.n_plots,
                                figsize=figsize, squeeze=False)

        for plotter, ax_row in zip(self.event_plotter_list, axs):
            major_label_set = False
            for view, ax in zip(self.views, ax_row):
                plotter.plot(view=view, fig=fig, ax=ax)

                if not major_label_set:
                    ylabel = ax.get_ylabel()
                    row_name = plotter.name.replace("_", "\\ ")
                    new_label = r"$\bf{" + row_name + "}$" + "\n" + ylabel
                    ax.set_ylabel(new_label)

                major_label_set = True

        fig.tight_layout()

    def set_same_scale(self):
        """
        Uses minimum and maximum of each dataset to find global min/max
        """
        for view in self.views:

            if not view.same_scale:
                # View controls same scale or not
                continue

            axis1_mins = []
            axis1_maxs = []

            axis2_mins = []
            axis2_maxs = []

            for plotter in self.event_plotter_list:
                lim_min, lim_max = plotter.get_view_limits_axis1(view)
                axis1_mins.append(lim_min)
                axis1_maxs.append(lim_max)

                lim_min, lim_max = plotter.get_view_limits_axis2(view)
                axis2_mins.append(lim_min)
                axis2_maxs.append(lim_max)

            view.set_axis1_limits(np.nanmin(axis1_mins), np.nanmax(axis1_maxs))
            if view.axis2 is not None:
                view.set_axis2_limits(np.nanmin(axis2_mins), np.nanmax(axis2_maxs))